/**
 * Max Fighter
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://maxfighter.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
#include <mysdl/math.h>
#include <mysdl/mysdl.h>

#include <mysdl/gui/text.h>

#include "maxfighter.h"
#include "game.h"
#include "resources.h"

namespace maxfighter
{
	#define min(x,y) ((x<y) ? (x) : (y))
	#define max(x,y) ((x>y) ? (x) : (y))
	#define sqr(x) ((x)*(x))
	#define sgn(x) ((x)==0 ? 0 : (x)<0 ? -1 : +1)

	/** Previously entered names of the two players. Used to provide the
	 *  old names as defaults in further games. */
	std::string previousName[2];

	const mysdl::Color white(1.0,1.0,1.0,1.0);
	const mysdl::Color red(1.0,0.0,0.0,1.0);
	const mysdl::Color blue(0.0,0.0,1.0,1.0);

	// ==================================================================
	//  Common functions
	// ==================================================================

	/**
	 * Create a one-dimensional sound vector from a 2D screen position
	 * vector. A return value of 0 represents the center, negative values
	 * a position to the left and positive values a position to the right
	 * of the listener.
	 */
	GLfloat	makeSoundVector(const mysdl::Vector2D &pos)
	{
		static GLfloat xc=resources->vpPlayfield.dimensions.x/2;

		return((pos.x-xc)/resources->vpPlayfield.dimensions.x);
	}

	// ==================================================================
	//  MaxfighterSprite
	// ==================================================================

	/*
	 * See class definition.
	 */
	MaxfighterSprite::MaxfighterSprite() : mysdl::sprites::ImageSprite(),deleted(false)
	{
	}

	/*
	 * See class definition.
	 */
	MaxfighterSprite::~MaxfighterSprite()
	{
	}

	/*
	* See class definition.
	*/
	void MaxfighterSprite::deleteMe()
	{
		if(!deleted)
		{
			deleted=true;
			deleteQueue.push_back(this);
		}
	}

	// ==================================================================
	//  ExplosionSprite
	// ==================================================================

	/*
	* See class definition.
	*/
	ExplosionSprite::ExplosionSprite(mysdl::gfx::Animation *animation,bool fixed) :
		MaxfighterSprite(),
		fixed(fixed),velocity(),anim(0)
	{
		blendMode=mysdl::bmAdd;
		anim=setAnimation(animation);
	}

	/*
	* See class definition.
	*/
	void ExplosionSprite::perform(GLfloat ticks)
	{
		mysdl::sprites::ImageSprite::perform(ticks);

		if(!fixed)
		{
			pos.x+=velocity.x*game->speed*ticks;
			pos.y+=velocity.y*game->speed*ticks;
		}

		if(anim->isFinished())
		{
			deleteMe();
		}
	}

	// ==================================================================
	//  Player
	// ==================================================================

	/*
	* See class definition.
	*/
	Player::Player(PlayerType type) :
			MaxfighterSprite(),
	alive(true),engine(0),shield(0),score(0),nextContainer(specs.containerScoreStep),
	laserDelay(0),laserSpeed(0),laserLevel(0),speed(0),
	speedupCount(0),fireRateCount(0),
	laserTimer(0),laserTimerSecondary(0),
	anim(0),
	shieldEnergy(0),timeWarpLevel(0),shielding(false),
	timeWarpDuration(0),
	hud()
	{
		blendMode=mysdl::bmNormal;

		anim=setAnimation(resources->animPlayer[type==ptLeft ? 1 : 0]);
		anim->target=specs.playerTiltAnimFrames;

		xMin=specs.playfieldBorderX;
		xMax=resources->vpPlayfield.dimensions.x-specs.playfieldBorderX;

		switch(type)
		{
			case ptSingle:
			{
				pos.x=(xMin+xMax)/2;
				break;
			}

			case ptLeft:
			{
				GLfloat x=xMin+xMax/2;
				pos.x=x*0.5;
				break;
			}

			case ptRight:
			{
				GLfloat x=xMin+xMax/2;
				pos.x=x*1.5;
				break;
			}
		}

		pos.y=specs.startY;

		visible=true;
		canCollide=true;

		game->activePlayers++;

		engine=new mysdl::sprites::ImageSprite();
		engine->setAnimation(resources->animEngine);
		engine->blendMode=mysdl::bmAdd;
		engineStutter=false;
		engine->visible=true;

		game->spritesPlayer.addFront(engine);

		shield=new Shield(this);
		game->spritesPlayer.addBack(shield);

		setSpeedupCount(0);
		setFireRateCount(0);
		laserLevel=0;

		laserTimer=0;
		laserTimerSecondary=0;
	}

	/*
	* See class definition.
	*/
	Player::~Player()
	{
		if(engine)
		{
			delete engine;
			engine=0;
		}
	}

	/*
	* See class definition.
	*/
	void Player::startTimeWarp()
	{
		if((alive)&&(timeWarpLevel>0))
		{
			timeWarpLevel--;
			hud.timewarp=timeWarpLevel;

			game->timeWarpFactorDestination=specs.timeWarpFactor;
			game->timeWarpDuration=timeWarpDuration=specs.timeWarpDuration;

			if(resources->settings.volumeSfx>0)
			{
				resources->sourcesGeneral->play(resources->sampleTimeWarp);
			}
		}
	}

	/*
	* See class definition.
	*/
	void Player::addScore(Uint32 add)
	{
		if(alive)
		{
			score+=add;

			if(score>nextContainer)
			{
				game->containerQueue++;
				nextContainer+=specs.containerScoreStep;
			}
		}
	}

	/*
	* See class definition.
	*/
	void Player::explode()
	{
		if(alive)
		{
			ExplosionSprite *ex=new ExplosionSprite(resources->animExplosionSlow,true);
			ex->visible=true;
			ex->pos=this->pos;
			owner->addBack(ex);

			setShield(false);
			visible=false;
			canCollide=false;
			engine->visible=false;
			shield->visible=false;
			direction.x=0;
			direction.y=0;
			alive=false;
			game->activePlayers--;

			if(resources->settings.volumeSfx>0)
			{
				mysdl::audio::Sample *sample=resources->sampleExplosion;
 				mysdl::audio::SoundProperties p=sample->properties;
 				p.setStereoPosition(makeSoundVector(pos));
 				resources->sourcesExplosions->play(sample,&p);
 			}

			if(game->activePlayers==0)
			{
				if((resources->musicManager)&&(resources->settings.volumeMusic>0))
				{
					resources->musicManager->clearQueue();
					resources->musicManager->addAsLast(resources->musicDeath);
					resources->musicManager->skipToNext();
				}
			}
		}
	}

	/*
	* See class definition.
	*/
	void Player::setShield(bool shield)
	{
		if((alive)&&(shield!=shielding))
		{
			if(!shield)
			{
				this->shield->released=true;
			}

			if((!shield)||(shieldEnergy>0))
			{
				shielding=shield;
			}
		}
	}

	/*
	* See class definition.
	*/
	void Player::setDirection(mysdl::Vector2D direction)
	{
		this->direction=direction;

		anim->target=(direction.x==0 ? specs.playerTiltAnimFrames :
																	(Uint32)(specs.playerTiltAnimFrames+((GLfloat)specs.playerTiltAnimFrames*direction.x)));
	}

	/*
	* See class definition.
	*/
	void Player::fire()
	{
		if((alive)&&(game->fireEnabled))
		{
			bool fired=false;

			if(laserTimer==0)
			{
				Laser *laser1=new Laser(resources->imgLaserRay,this);
				Laser *laser2=new Laser(resources->imgLaserRay,this);

				laser1->pos.x-=(laserLevel==0 ? 15 : (laserLevel==1 ? 25 : 7));
				laser2->pos.x+=(laserLevel==0 ? 15 : (laserLevel==1 ? 25 : 7));

				laser1->velocity=mysdl::Vector2D(0.0,1.0)*laserSpeed;
				laser2->velocity=mysdl::Vector2D(0.0,1.0)*laserSpeed;

				game->spritesLaser.addFront(laser1);
				game->spritesLaser.addFront(laser2);

				fired=true;
				laserTimer=laserDelay;
			}

			if(laserTimerSecondary==0)
			{
				if(laserLevel>=2)
				{
					Laser *laser1=new Laser(resources->imgLaserRound,this);
					Laser *laser2=new Laser(resources->imgLaserRound,this);

					laser1->pos.x-=25;
					laser2->pos.x+=25;

					laser1->pos.y-=10;
					laser2->pos.y-=10;

					if(laserLevel==2)
					{
						laser1->velocity=mysdl::Vector2D(-0.05,+0.95)*laserSpeed;
						laser2->velocity=mysdl::Vector2D(+0.05,+0.95)*laserSpeed;
					}
					else
					{
						laser1->velocity=mysdl::Vector2D(-0.08,+0.92)*laserSpeed;
						laser2->velocity=mysdl::Vector2D(+0.08,+0.92)*laserSpeed;
					}

					game->spritesLaser.addFront(laser1);
					game->spritesLaser.addFront(laser2);

					if(laserLevel>3)
					{
						Laser *laser1=new Laser(resources->imgLaserRound,this);
						Laser *laser2=new Laser(resources->imgLaserRound,this);

						laser1->pos.x-=35;
						laser1->pos.y-=15;
						laser2->pos.x+=35;
						laser2->pos.y-=15;

						laser1->velocity=mysdl::Vector2D(-0.90,-0.10)*laserSpeed;
						laser2->velocity=mysdl::Vector2D(+0.90,-0.10)*laserSpeed;

						game->spritesLaser.addFront(laser1);
						game->spritesLaser.addFront(laser2);
					}

					fired=true;

					laserTimerSecondary=laserDelay*(laserLevel==2 ? 1.5 : (laserLevel==3 ? 1.3 : 1.1));
				}
			}

			if(fired)
			{
				if(resources->settings.volumeSfx>0)
				{
					mysdl::audio::Sample *sample=resources->sampleLaser;
					mysdl::audio::SoundProperties p=sample->properties;
					p.setStereoPosition(makeSoundVector(pos));
					resources->sourcesExplosions->play(sample,&p);
				}
			}
		}
	}

	/*
	* See class definition.
	*/
	void Player::perform(GLfloat ticks)
	{
		if(timeWarpDuration>0)
		{
			ticks/=game->timeWarpFactor;
		}

		Uint32 intTicks=(Uint32)ticks;

		if(timeWarpDuration>0)
		{
			if(timeWarpDuration>intTicks)
			{
				timeWarpDuration-=intTicks;
			}
			else
			{
				timeWarpDuration=0;
			}
		}

		if(shield->color.a>0)
		{
			GLfloat shieldConsumption=ticks*specs.shieldConsumption;

			if(shieldEnergy>=shieldConsumption)
			{
				shieldEnergy-=shieldConsumption;
			}
			else
			{
				shieldEnergy=0;
				setShield(false);
			}
		}

		if((shielding)&&(!shield->collapsing))
		{
			if(!shield->visible)
			{
				shield->visible=true;
				shield->color.a=0;
			}

			if(shield->color.a<1)
			{
				shield->color.a+=specs.shieldFadeInFactor*ticks;

				shield->canCollide=shield->color.a>=0.5;

				if(shield->color.a>1)
				{
					shield->color.a=1;
				}
			}
		}
		else
		{
			if(shield->color.a>0)
			{
				shield->color.a-=(shield->collapsing ? specs.shieldCollapseFactor : specs.shieldFadeOutFactor)*ticks;

				shield->canCollide=shield->color.a>=0.5;

				if(shield->color.a<=0)
				{
					shield->color.a=0;
					shield->visible=false;
				}
			}

			if((shield->color.a==0)&&(shield->released))
			{
				shield->released=false;
				shield->collapsing=false;
			}
		}

		MaxfighterSprite::perform(ticks);

		mysdl::Vector2D newPos=pos+direction*ticks*speed;

		if(newPos.x<xMin)
		{
			newPos.x=xMin;
		}
		else if(newPos.x>=xMax)
		{
			newPos.x=xMax;
		}

		GLfloat maxY=(game->type==gtSingleB ? specs.maxYModeB : specs.maxYModeA);

		if(newPos.y<specs.standardY)
		{
			newPos.y=specs.standardY;
		}
		else if(newPos.y>specs.standardY+maxY)
		{
			newPos.y=specs.standardY+maxY;
		}

		if(game->activePlayers>1)
		{
			Player *p=game->players[this==game->players[1] ? 0 : 1];

			int dx=abs((int)(newPos.x-p->pos.x));
			int dy=abs((int)(newPos.y-p->pos.y));

			if((dx<specs.minPlayerDistanceX)&&(dy<specs.minPlayerDistanceY))
			{
				if(dx<dy)
				{
					newPos.y=pos.y;
				}
				else
				{
					newPos.x=pos.x;
				}
			}

			dx=abs((int)(newPos.x-p->pos.x));
			dy=abs((int)(newPos.y-p->pos.y));

			if((dx<specs.minPlayerDistanceX)&&(dy<specs.minPlayerDistanceY))
			{
				newPos.y=pos.y;
				newPos.x=pos.x;
			}
		}

		pos=newPos;

		if(laserTimer<ticks)
		{
			laserTimer=0;
		}
		else
		{
			laserTimer-=ticks;
		}

		if(laserTimerSecondary<ticks)
		{
			laserTimerSecondary=0;
		}
		else
		{
			laserTimerSecondary-=ticks;
		}

		engine->pos=pos;
		shield->pos=pos;
	}

	/*
	* See class definition.
	*/
	void Player::setSpeedupCount(int speedupCount)
	{
		this->speedupCount=speedupCount;
		this->speed=specs.speedStart+(GLfloat)speedupCount*specs.speedIncrement;
	}

	/*
	* See class definition.
	*/
	void Player::setFireRateCount(int fireRateCount)
	{
		this->fireRateCount=fireRateCount;
		laserDelay=specs.getLaserDelay(fireRateCount);
		laserSpeed=specs.getLaserSpeed(fireRateCount);
	}

	/*
	* See class definition.
	*/
	void Player::updateHud(bool startTimer)
	{
		hud.fireRate=fireRateCount;
		hud.fireType=laserLevel;
		hud.speedup=speedupCount;
		hud.timewarp=timeWarpLevel;

		if(startTimer)
		{
			hud.switchCountdown=specs.hudSwitchTime;
			hud.switchesLeft=specs.hudSwitchCount;
			hud.actual=true;
		}
	}

	/*
	* See class definition.
	*/
	bool Player::collectExtra(Extra *extra)
	{
		if(extra->collected)
		{
			return false;
		}

		bool unstutter=false;

		extra->collected=true;

		switch(extra->extraType)
		{
			case etSpeedup:
			{
				updateHud(true);

				if(speedupCount<specs.speedupCountMax)
				{
					setSpeedupCount(speedupCount+1);
				}

				unstutter=true;

				break;
			}

			case etFireRate:
			{
				updateHud(true);

				if(fireRateCount<specs.fireRateCountMax)
				{
					setFireRateCount(fireRateCount+1);
				}

				unstutter=true;

				break;
			}

			case etLaserType:
			{
				updateHud(true);

				if(laserLevel<specs.laserLevelMax)
				{
					laserLevel++;
				}

				unstutter=true;

				break;
			}

			case etShield:
			{
				shieldEnergy+=specs.shieldEnergyPacket;

				if(shieldEnergy>specs.shieldEnergyMax)
				{
					shieldEnergy=specs.shieldEnergyMax;
				}

				break;
			}

			case etTimeWarp:
			{
				updateHud(true);

				if(timeWarpLevel<specs.timeWarpLevelMax)
				{
					timeWarpLevel++;
				}

				break;
			}

			case etScore:
			{
				addScore(specs.scoreCrystal);
				break;
			}

			default:
				assert(0);
		}

		if(unstutter)
		{
			if(engineStutter)
			{
				engine->setAnimation(resources->animEngine);
				engineStutter=false;
			}
		}

		if(resources->settings.volumeSfx>0)
		{
			mysdl::audio::Sample *sample=resources->sampleExtra;
			mysdl::audio::SoundProperties p=sample->properties;
			p.setStereoPosition(makeSoundVector(pos));
			resources->sourcesExplosions->play(sample,&p);
		}

		return(true);
	}

	/*
	* See class definition.
	*/
	void Player::onCollision(mysdl::sprites::Sprite *spr)
	{
		if(Extra *extra=dynamic_cast<Extra *>(spr))
		{
			collectExtra(extra);
		}
		else
		{
			if(!(shield->canCollide))
			{
				hit(dynamic_cast<MaxfighterSprite *>(spr));

				if(Laser *laser=dynamic_cast<Laser *>(spr))
				{
					laser->deleteMe();
				}
			}
		}
	}

	/*
	* See class definition.
	*/
	void Player::hit(MaxfighterSprite *sprite)
	{
		int damage=(dynamic_cast<Foe *>(sprite)) ? 2 : 1;

		if((speedupCount<damage)&&(fireRateCount<damage)&&(laserLevel<damage))
		{
			explode();
		}
		else
		{
			if(sprite)
			{
				if(Asteroid *asteroid=dynamic_cast<Asteroid *>(sprite)) // TODO: unify?
				{
					asteroid->explode(false);
				}
				else if(Container *container=dynamic_cast<Container *>(sprite))
				{
					container->explode();
				}
				else if(Foe *foe=dynamic_cast<Foe *>(sprite))
				{
					foe->explode();
				}
			}

			updateHud(true);

			if(speedupCount>0)
			{
				setSpeedupCount(mysdl::max<int>(0,speedupCount-damage));

				Extra *e=new Extra(etSpeedup,pos);
				e->canCollide=false;
				e->ticks=1;
				e->velocity=mysdl::Vector2D(-0.0,-1.0);
				e->pos.y-=15;

				game->spritesExtras.addBack(e);
			}

			if(fireRateCount>0)
			{
				setFireRateCount(mysdl::max<int>(0,fireRateCount-damage));

				Extra *e=new Extra(etFireRate,pos);
				e->canCollide=false;
				e->ticks=1;
				e->velocity=mysdl::Vector2D(-0.3,-0.7);
				e->pos.x-=10;
				e->pos.y-=15;

				game->spritesExtras.addBack(e);
			}

			if(laserLevel>0)
			{
				laserLevel=mysdl::max<int>(0,laserLevel-damage);

				Extra *e=new Extra(etLaserType,pos);
				e->canCollide=false;
				e->ticks=1;
				e->velocity=mysdl::Vector2D(+0.3,-0.7);
				e->pos.x+=10;
				e->pos.y-=15;

				game->spritesExtras.addBack(e);
			}

			if((speedupCount==0)&&(fireRateCount==0)&&(laserLevel==0))
			{
				engine->setAnimation(resources->animEngineStutter);
				engineStutter=true;
			}
		}
	}

	// ==================================================================
	//  Shield
	// ==================================================================

	/*
	* See class definition.
	*/
	Shield::Shield(Player *player) : ImageSprite(),player(player),collapsing(false),released(false)
	{
		setAnimation(resources->animShield);
		blendMode=mysdl::bmAdd;
		visible=false;
		color.a=0.0;
		canCollide=false;
	}

	/*
	* See class definition.
	*/
	void Shield::onCollision(mysdl::sprites::Sprite *sprite)
	{
		if(Asteroid *asteroid=dynamic_cast<Asteroid *>(sprite))
		{
			asteroid->explode(true);
		}
		else if(Container *container=dynamic_cast<Container *>(sprite))
		{
			container->explode();
		}
		else if(Laser *laser=dynamic_cast<Laser *>(sprite))
		{
			laser->deleteMe();
		}
	}

	// ==================================================================
	//  Laser
	// ==================================================================

	/*
	* See class definition.
	*/
	Laser::Laser(mysdl::gfx::Image *img,MaxfighterSprite *source,GLfloat ttl,GLfloat fadeSpeed) : MaxfighterSprite(),source(source),velocity(0,0),ttl(ttl),fadeSpeed(fadeSpeed)
	{
		setImage(img);

		this->pos=source->pos;
		visible=true;
		canCollide=true;
	}

	/*
	* See class definition.
	*/
	void Laser::perform(GLfloat ticks)
	{
		pos+=velocity*ticks;

		if(ttl>0)
		{
			ttl-=ticks;

			if(ttl<=0)
			{
				ttl=0;
			}
		}

		if(ttl==0)
		{
			color.a-=fadeSpeed*ticks;

			if(color.a<=0)
			{
				color.a=0;
				deleteMe();
			}
			else if(color.a<0.5)
			{
				canCollide=false;
			}
		}

		MaxfighterSprite::perform(ticks);

		if((pos.y>resources->vpPlayfield.dimensions.y+dimensions.y-hotspot.y)||
			 (pos.x<-dimensions.x+hotspot.x)||
			 (pos.x>resources->vpPlayfield.dimensions.x+hotspot.x))
		{
			deleteMe();
		}
	}

	// ==================================================================
	//  Asteroid
	// ==================================================================

	/*
	* See class definition.
	*/
	Asteroid::Asteroid(AsteroidType type,int energy,bool regular) :
		MaxfighterSprite(),
		type(type),energy(energy),regular(regular),
		velocity(),destroy(false)//,velocityModifier()
	{
		if(regular)
		{
			game->asteroidCount++;
		}

		if(type==atSmall)
		{
			setAnimation(resources->animAsteroidSmall,mysdl::random<GLfloat>(0.5,1.5));
		}
		else
		{
			setAnimation(resources->animAsteroidLarge,mysdl::random<GLfloat>(0.5,1.5));
		}

		setColor(mysdl::Color(1.0-0.1*(GLfloat)energy,1.0-0.1*(GLfloat)energy,1.0,1.0));

		visible=true;
		canCollide=true;
	}

	/*
	* See class definition.
	*/
	Asteroid::~Asteroid()
	{
		if(regular)
		{
			game->asteroidCount--;
		}
	}

	/*
	* See class definition.
	*/
	void Asteroid::onCollision(mysdl::sprites::Sprite *sprite)
	{
		if(Laser *laser=dynamic_cast<Laser *>(sprite))
		{
			if(energy>0)
			{
				energy--;
				setColor(mysdl::Color(1.0-0.1*(mysdl::myfloat)energy,1.0-0.1*(mysdl::myfloat)energy,1.0,1.0));
			}
			else
			{
				Player *player=0;

				if(laser->source==game->players[0])
				{
					player=game->players[0];
				}
				else if(laser->source==game->players[1])
				{
					player=game->players[1];
				}

				game->destroyedAsteroid(player,this);
				explode(false);
			}

			laser->deleteMe();
		}
	}

	/*
	* See class definition.
	*/
	void Asteroid::explode(bool complete)
	{
		if(visible)
		{
			if((type==atLarge)&&(!complete))
			{
				Asteroid *a1=new Asteroid(atSmall,energy,false);
				Asteroid *a2=new Asteroid(atSmall,energy,false);

				a1->pos=a2->pos=pos;
				a1->velocity=a2->velocity=velocity;

				a1->velocity.x-=specs.splitSpeedX;
				a2->velocity.x+=specs.splitSpeedX;

				game->spritesAsteroids.addFront(a1);
				game->spritesAsteroids.addFront(a2);
			}

			ExplosionSprite *ex=new ExplosionSprite(resources->animExplosionFast);
			ex->visible=true;
			ex->pos=this->pos;
			ex->velocity=this->velocity;
			ex->velocity.y-=game->speed;
			owner->addBack(ex);

			if(resources->settings.volumeSfx>0)
			{
				mysdl::audio::Sample *sample=resources->sampleExplosion;
				mysdl::audio::SoundProperties p=sample->properties;
				p.setStereoPosition(makeSoundVector(pos));
				resources->sourcesExplosions->play(sample,&p);
			}

			destroy=true;
		}
	}

	/*
	* See class definition.
	*/
	void Asteroid::perform(GLfloat ticks)
	{
		MaxfighterSprite::perform(ticks);

		if(destroy)
		{
			visible=false;
			canCollide=false;
			deleteMe();
		}
		else
		{
			pos+=velocity*ticks;

			if((pos.y<-hotspot.y)||
				(pos.x<-dimensions.x+hotspot.x)||
				(pos.x>resources->vpPlayfield.dimensions.x+hotspot.x))
			{
				deleteMe();
			}
		}
	}

	// ==================================================================
	//  Container
	// ==================================================================

	/*
	* See class definition.
	*/
	Container::Container() :
		MaxfighterSprite(),extraType(etUnknown)
	{
		extraType=game->getExtraType();

		int animIndex;

		switch(extraType)
		{
			case etShield:
			{
				animIndex=1;
				break;
			}

			case etTimeWarp:
			{
				animIndex=2;
				break;
			}

			default:
			{
				animIndex=0;
			}
		}

		setAnimation(resources->animContainer[animIndex]);

		pos.x=mysdl::random<GLfloat>(specs.playfieldBorderX,resources->vpPlayfield.dimensions.x-specs.playfieldBorderX);
		pos.y=resources->vpPlayfield.dimensions.y+50;

		visible=true;
		canCollide=true;
	}

	/*
	* See class definition.
	*/
	void Container::onCollision(mysdl::sprites::Sprite *sprite)
	{
		if(Laser *laser=dynamic_cast<Laser *>(sprite))
		{
			explode();
			laser->deleteMe(); // TODO: is this how it works with asteroids?
		}
	}

	/*
	* See class definition.
	*/
	void Container::explode()
	{
		if(!deleted)
		{
			Extra *extra=new Extra(extraType,pos);
			game->spritesExtras.addFront(extra);

			ExplosionSprite *ex=new ExplosionSprite(resources->animExplosionFast);
			ex->visible=true;
			ex->pos=this->pos;
			ex->velocity.y=-specs.containerSpeedFactor*specs.baseSpeed;
			owner->addBack(ex);

			if(resources->settings.volumeSfx>0)
			{
				mysdl::audio::Sample *sample=resources->sampleExplosion;
				mysdl::audio::SoundProperties p=sample->properties;
				p.setStereoPosition(makeSoundVector(pos));
				resources->sourcesExplosions->play(sample,&p);
			}

			deleteMe();
			visible=false;
		}
	}

	/*
	* See class definition.
	*/
	void Container::perform(GLfloat ticks)
	{
		MaxfighterSprite::perform(ticks);

		pos.y-=specs.containerSpeedFactor*specs.baseSpeed*ticks;

		if(pos.y<-50)
		{
			deleteMe();
		}
	}

	// ==================================================================
	//  Extra
	// ==================================================================

	/*
	* See class definition.
	*/
	Extra::Extra(ExtraType extraType,mysdl::Vector2D pos) : MaxfighterSprite(),extraType(extraType),collected(false),velocity(),ticks(0)
	{
		blendMode=mysdl::bmAdd;

		if(extraType==etSpeedup)
		{
			setAnimation(resources->animExtraSpeedup);
		}
		else if(extraType==etFireRate)
		{
			setAnimation(resources->animExtraFireRate);
		}
		else if(extraType==etLaserType)
		{
			setAnimation(resources->animExtraLaserType);
		}
		else if(extraType==etShield)
		{
			setAnimation(resources->animExtraShield);
		}
		else if(extraType==etTimeWarp)
		{
			setAnimation(resources->animExtraTimeWarp);
		}
		else
		{
			setAnimation(resources->animExtraScore);
		}

		this->pos=pos;

		visible=true;
		canCollide=true;
	}

	/*
	* See class definition.
	*/
	void Extra::perform(GLfloat ticks)
	{
		MaxfighterSprite::perform(ticks);

		if(this->ticks>0)
		{
			this->ticks+=(Uint32)ticks;

			if(this->ticks>=specs.extraReviveTicks)
			{
				this->ticks=0;
				canCollide=true;
			}
		}

		if(velocity.y!=0)
		{
			pos+=velocity*ticks*specs.extraVelocityFactor;
		}
		else
		{
			pos.y-=specs.extraSpeedFactor*game->speed*ticks;
		}

		if(pos.y<-50)
		{
			deleteMe();
		}
		else if(collected)
		{
			color-=color*(specs.extraFadeFactor*ticks);

			if(color.a<0)
			{
				deleteMe();
			}
		}
	}

	// ==================================================================
	//  Foe
	// ==================================================================

	/*
	* See class definition.
	*/
	Foe::Foe(bool right) :
		MaxfighterSprite(),
		fireDelay(0),volleyCount(1),
		acceleration(0),distanceSpeedFactor(0),floorCountdown(0),kamikaze(true),
		right(right),velocity(-1),anim(0),volley(0),laserTimer(0),alive(true),hull(0),leaving(false)
	{
		hull=specs.getFoeHull(game->level);
		fireDelay=specs.getFoeFireDelay(game->level);
		volleyCount=specs.getFoeVolleyCount(game->level);
		volleyDelay=specs.getFoeVolleyDelay(game->level);

		velocity=specs.foeSpeed;
		acceleration=specs.getFoeAcceleration(game->level);
		distanceSpeedFactor=specs.getFoeDistanceSpeedFactor(game->level);

		anim=setAnimation(right ? resources->animFoeRight : resources->animFoeLeft);

		pos.x=(right ? 770 : 30);
		pos.y=850;

		visible=true;
		canCollide=true;

		game->spritesFoes.addBack(this);
	}

	/*
	* See class definition.
	*/
	Foe::~Foe()
	{
		int i=right ? 1 : 0;

		if(game->foes[i]==this)
		{
			game->foes[i]=0;
		}
	}

	/*
	* See class definition.
	*/
	void Foe::perform(GLfloat ticks)
	{
		MaxfighterSprite::perform(ticks);

		if(alive)
		{
			if(game->activePlayers==0)
			{
				leaving=true;
				velocity=-1.0;
			}

			if(!leaving)
			{
				if(laserTimer>0)
				{
					if(laserTimer>ticks)
					{
						laserTimer-=ticks;
					}
					else
					{
						laserTimer=0;
					}
				}

				Player *player=game->players[0];
				GLfloat dx=player->pos.x-pos.x; // FIXME: may be dead

				if(dx<0)
				{
					dx=-dx;
				}

				if(game->players[1])
				{
					GLfloat dx2=game->players[1]->pos.x-pos.x; // FIXME: may be dead

					if(dx2<0)
					{
						dx2=-dx2;
					}

					if(dx2<dx)
					{
						dx=dx2;
						player=game->players[1];
					}
				}

				if((floorCountdown<=0)&&(mysdl::random(specs.getFoeFloorChance(game->level))<1))
				{
					floorCountdown=specs.getFoeFloorDuration(game->level);
				}

				if((dx<120)&&(!kamikaze)&&(mysdl::random(specs.getFoeKamikazeChance(game->level))<1))
				{
					kamikaze=true;
				}

				GLfloat targetY;

				if(floorCountdown>0)
				{
					targetY=specs.foeFloor;
				}
				else if((dx<90)&&((!kamikaze)||(player->shielding)))
				{
					targetY=specs.foeTop;
				}
				else
				{
					targetY=player->pos.y;
				}

				float dy=targetY-pos.y;
				float targetVelocity=sgn(dy)*(dy*dy)/distanceSpeedFactor;

				if((kamikaze)&&(targetVelocity>0))
				{
					kamikaze=false;
				}

				if(velocity<targetVelocity)
				{
					velocity+=ticks*acceleration;

					if(velocity>targetVelocity)
					{
						velocity=targetVelocity;
					}
				}
				else if(velocity>targetVelocity)
				{
					velocity-=ticks*acceleration;

					if(velocity<targetVelocity)
					{
						velocity=targetVelocity;
					}
				}

				if(velocity<-1)
				{
					velocity=-1;
				}
				else if(velocity>1)
				{
					velocity=1;
				}

				if(floorCountdown>0)
				{
					if(floorCountdown>(Uint32)ticks)
					{
						floorCountdown-=(Uint32)ticks;
					}
					else
					{
						floorCountdown=0;
					}
				}

				if(!kamikaze)
				{
					fire();
				}
			}

			anim->target=(velocity==0 ? 5 : velocity<0 ? 5-(int)(5.0*-velocity+0.5) : 6+(int)(5.0*velocity+0.5));

			pos.y+=velocity*specs.foeSpeed*ticks;

			if((leaving)&&(pos.y<-50))
			{
				deleteMe();
			}
		}
	}

	/*
	* See class definition.
	*/
	void Foe::fire()
	{
		if(laserTimer<=0)
		{
			Laser *laser0=new Laser(resources->/*anim*/imgLaserFoe,this,specs.getFoeLaserTtl(game->level),specs.foeLaserFadeOutSpeed);

			laser0->pos.x+=(right ? -20 : +20);
			laser0->velocity=mysdl::Vector2D((right ? -1.0 : +1.0)*specs.foeLaserSpeed,velocity*specs.foeLaserSpeedVertical);

			game->spritesFoeLaser.addFront(laser0);

			if(resources->settings.volumeSfx>0)
			{
				mysdl::audio::Sample *sample=resources->sampleLaser;
				mysdl::audio::SoundProperties p=sample->properties;
				p.setStereoPosition(makeSoundVector(pos));
				resources->sourcesExplosions->play(sample,&p);
			}

			if(volley)
			{
				volley--;
			}
			else
			{
				volley=volleyCount;
			}

			laserTimer=(volley ? volleyDelay : fireDelay);
		}
	}

	/*
	* See class definition.
	*/
	void Foe::onCollision(mysdl::sprites::Sprite *sprite)
	{
		MaxfighterSprite *source=0;

		if(Laser *laser=dynamic_cast<Laser *>(sprite))
		{
			source=laser->source;
			laser->deleteMe();
		}

		if(source)
		{
			hull--;

			if(hull==0)
			{
				Player *player=0;

				if(source==game->players[0])
				{
					player=game->players[0];
				}
				else if(source==game->players[1])
				{
					player=game->players[1];
				}

				game->destroyedFoe(player,this);

				explode();
			}
		}
	}

	/*
	* See class definition.
	*/
	void Foe::explode()
	{
		if(alive)
		{
			ExplosionSprite *ex=new ExplosionSprite(resources->animExplosionSlow,true);
			ex->visible=true;
			ex->pos=this->pos;

			owner->addBack(ex);

			visible=false;
			canCollide=false;
			alive=false;
			deleteMe();

			if(resources->settings.volumeSfx>0)
			{
				mysdl::audio::Sample *sample=resources->sampleExplosion;
				mysdl::audio::SoundProperties p=sample->properties;
				p.setStereoPosition(makeSoundVector(pos));
				resources->sourcesExplosions->play(sample,&p);
			}
		}
	}

	// ==================================================================
	//  Game
	// ==================================================================

	/*
	* See class definition.
	*/
	Game::Game(GameType type) :
			GameState(),
	evaluators(),axisPlayerX(),axisPlayerY(),triggerPlayerShield(),
	type(type),over(false),leaving(false),levelComplete(false),level(0),playerCount(0),activePlayers(0),
	spritesPlayer(),spritesLaser(),spritesAsteroids(),spritesExtras(),spritesFoes(),spritesFoeLaser(),
	players(),foes(),state(igsGameIntro),stateChangeCountdown(0),asteroidCount(0),
	interludeColor(0.5,0.5,1.0,0.0),
	highscoresChecked(0),fireEnabled(false),
	timeWarpDuration(0),timeWarpFactor(1),timeWarpFactorDestination(1),
	containerQueue(0),
	pingCount(0)
	{
		resources->initGame();
		game=this;

		init();
	}

	/*
	* See class definition.
	*/
	Game::~Game()
	{
		uninit();

		resources->uninitGame();
		deactivate();
		game=0;
	}

	/*
	* See class definition.
	*/
	void Game::activate()
	{
		GameState::activate();

		resources->evaluatorGameCore->enabled=true;

		evaluators[0]->enabled=true;

		if(evaluators[1])
		{
			evaluators[1]->enabled=true;
		}

		resources->evaluatorMenu->enabled=false;
		resources->gui->Evaluator::enabled=false;
	}

	/*
	* See class definition.
	*/
	void Game::deactivate()
	{
		evaluators[0]->enabled=false;

		if(evaluators[1])
		{
			evaluators[1]->enabled=false;
		}

		resources->evaluatorGameCore->enabled=false;

		GameState::deactivate();

		menu->activate();
	}

	/*
	* See class definition.
	*/
	void Game::init()
	{
		terminated=false;

		for(int i=0; i<2; i++)
		{
			highscorePos[i]=-1;
		}

		spritesPlayer.clear();

		over=false;
		leaving=false;
		levelComplete=false;
		level=0;
		state=igsGameIntro;
		stateChangeCountdown=0;
		asteroidCount=0;
		fireEnabled=false;
		timeWarpDuration=0;
		timeWarpFactor=1;
		timeWarpFactorDestination=1;
		containerQueue=0;

		spritesLaser.clear();
		spritesAsteroids.clear();
		spritesExtras.clear();
		spritesFoes.clear();
		spritesFoeLaser.clear();

		playerCount=(type>=gtMulti ? 2 : 1);
		players[0]=players[1]=0;
		foes[0]=foes[1]=0;

		asteroidTimer=0;
		speedupTimer=0;
		extraTimerScore=specs.extraIntervalScore/2;
		extraTimerSpeedup=0;

		switch(type)
		{
			case gtMulti:
				evaluators[0]=resources->evaluatorPlayerMulti[0];
				evaluators[1]=resources->evaluatorPlayerMulti[1];
				break;

			default:
				evaluators[0]=resources->evaluatorPlayerSingle[0];
				evaluators[1]=0;
		}

		axisPlayerX[0]=(*evaluators[0])["x"];
		axisPlayerY[0]=(*evaluators[0])["y"];
		triggerPlayerShield[0]=(*evaluators[0])["shield"];

		if(playerCount>1)
		{
			axisPlayerX[1]=(*evaluators[1])["x"];
			axisPlayerY[1]=(*evaluators[1])["y"];
			triggerPlayerShield[1]=(*evaluators[1])["shield"];
		}

		players[0]=new Player(type>=gtMulti ? ptRight : ptSingle);
		spritesPlayer.addFront(players[0]);

		if(type>=gtMulti)
		{
			players[1]=new Player(ptLeft);
			spritesPlayer.addFront(players[1]);
		}

		spritesExtras.blendMode=mysdl::bmAdd;

		initLevel(1);

		startMusic();

		resources->queue->flush();
	}

	/*
	* See class definition.
	*/
	void Game::uninit()
	{
		if(resources->musicManager)
		{
			resources->musicManager->stop();
		}

		for(Uint32 i=0; i<playerCount; i++)
		{
			if(players[i])
			{
				delete players[i];
				players[i]=0;
			}
		}
	}

	/*
	* See class definition.
	*/
	void Game::restart()
	{
		uninit();
		init();
	}

	/*
	* See class definition.
	*/
	void Game::closeWindowSoundSettings(bool confirm)
	{
		GameState::closeWindowSoundSettings(confirm);
		resources->paneInGame->setVisible(true);
		resources->btnInGameSoundSettings->setFocused(true);
	}

	/*
	* See class definition.
	*/
	void Game::closeWindowControls(bool confirm)
	{
		GameState::closeWindowControls(confirm);
		resources->paneInGame->setVisible(true);

		if(originalEvaluator==resources->evaluatorPlayerSingle[0])
		{
			resources->btnInGameControlsSinglePlayer->setFocused(true);
		}
		else if(originalEvaluator==resources->evaluatorPlayerMulti[0])
		{
			resources->btnInGameControlsMultiplayer[0]->setFocused(true);
		}
		else if(originalEvaluator==resources->evaluatorPlayerMulti[1])
		{
			resources->btnInGameControlsMultiplayer[1]->setFocused(true);
		}
	}

	/*
	* See class definition.
	*/
	void Game::handleEvent(const mysdl::Trigger *e)
	{
		switch(e->id)
		{
			case cgQuit:
			{
				if(state==igsEnterNames)
				{
					break;
				}

				resources->paneInGame->setVisible(false);
				resources->gui->setActive(false);
				resources->gui->setVisible(false);

				terminated=true;

				if((players[0])&&(players[0]->alive))
				{
					players[0]->explode();
				}

				if((players[1])&&(players[1]->alive))
				{
					players[1]->explode();
				}

				// begin duplicate code
				resources->paneInGame->setVisible(false);
				resources->gui->setActive(false);
				resources->gui->setVisible(false);
				paused=false;

				evaluators[0]->enabled=true;

				if(evaluators[1])
				{
					evaluators[1]->enabled=true;
				}

				resources->gui->Evaluator::enabled=false;
				// end duplicate code
				break;
			}

			case cgPause:
			{
				if(state==igsEnterNames)
				{
					break;
				}

				if(activePlayers>0)
				{
					if(!paused)
					{
						resources->paneInGame->setVisible(true);
						resources->gui->setActive(true);
						resources->gui->setVisible(true);
						resources->btnContinue->setFocused(true);
						paused=true;

						evaluators[0]->enabled=false;

						if(evaluators[1])
						{
							evaluators[1]->enabled=false;
						}

						resources->gui->Evaluator::enabled=true;
					}
				}
				
				break;
			}
			
			case cgExit:
			{
				if(closeWindows(false))
				{
					break;
				}

				if(state==igsEnterNames)
				{
					break;
				}

				if(activePlayers>0)
				{
					if(paused)
					{
						// begin duplicate code
						resources->paneInGame->setVisible(false);
						resources->gui->setActive(false);
						resources->gui->setVisible(false);
						paused=false;

						evaluators[0]->enabled=true;

						if(evaluators[1])
						{
							evaluators[1]->enabled=true;
						}

						resources->gui->Evaluator::enabled=false;
						resources->queue->flush();
						// end duplicate code
					}
					else
					{
						resources->paneInGame->setVisible(true);
						resources->gui->setActive(true);
						resources->gui->setVisible(true);
						resources->btnContinue->setFocused(true);
						paused=true;

						evaluators[0]->enabled=false;

						if(evaluators[1])
						{
							evaluators[1]->enabled=false;
						}

						resources->gui->Evaluator::enabled=true;
					}
				}
				else if((resources->musicManager)&&(resources->musicManager->isPlaying()))
				{
					leaving=true;
				}
				else
				{
					delete this;
					break;
				}

				break;
			}

			case cgOk:
			{
				if(closeWindows(true))
				{
					break;
				}

				if(state==igsEnterNames)
				{
					mysdl::gui::Textfield *tf=(mysdl::gui::Textfield *)((*resources->gui)["tfEnterName"]);
					std::string name=tf->getCaption();

					if(name.size()==0)
					{
						break;
					}

					previousName[highscoresChecked]=name;
					game->getHighscores()->add(HighscoreEntry(name,players[highscoresChecked]->score));

					((*resources->gui)["pEnterName"])->setVisible(false);
					resources->gui->setVisible(false);
					resources->gui->setVisible(false);
					resources->gui->setActive(false);

					evaluators[0]->enabled=true;

					if(evaluators[1])
					{
						evaluators[1]->enabled=true;
					}

					resources->gui->Evaluator::enabled=false;

					setHighscoresChecked(highscoresChecked+1);
				}
				else if(activePlayers==0) // TODO: use state
				{
					restart();
				}

				break;
			}

			case cgFire:
			{
				if(activePlayers>0)
				{
					if(!paused)
					{
						players[e->data]->fire();
					}
				}
				else if((state==igsShowHighscores)&&(highscoreTime>=specs.restartDelay))
				{
					restart();
				}

				break;
			}

			case cgTimeWarp:
			{
				if(!paused)
				{
					players[e->data]->startTimeWarp();
				}

				break;
			}

			default:
			{
				GameState::handleEvent(e);
			}
		}
	}

	/*
	* See class definition.
	*/
	ExtraType Game::getExtraType()
	{
		static GLfloat p[extraCount];
		static GLfloat remaining[extraCount];

		ExtraType extraType;

		for(unsigned int i=0; i<playerCount; i++)
		{
			if((players[i])&&(players[i]->alive))
			{
				if(type!=gtSingleB)
				{
					int extraLevel=min(min(players[i]->fireRateCount,players[i]->laserLevel),players[i]->speedupCount);

					remaining[etFireRate]=max(min((level/2)+1,specs.fireRateCountMax-specs.extraContainerLimit)-players[i]->fireRateCount,0);
					remaining[etLaserType]=max(min((level/2)+1,specs.laserLevelMax-specs.extraContainerLimit)-players[i]->laserLevel,0);
					remaining[etSpeedup]=max(min((level/2)+1,specs.speedupCountMax-specs.extraContainerLimit)-players[i]->speedupCount,0);
					remaining[etShield]=extraLevel<specs.minLevelForShield ? 0 : (specs.shieldEnergyMax-players[i]->shieldEnergy)/specs.shieldEnergyPacket;
					remaining[etTimeWarp]=extraLevel<specs.minLevelForWeapons ? 0 : specs.timeWarpLevelMax-players[i]->timeWarpLevel;
				}
				else
				{
					remaining[etFireRate]=0;
					remaining[etLaserType]=0;
					remaining[etSpeedup]=specs.speedupCountMax-players[i]->speedupCount;
					remaining[etShield]=0;
					remaining[etTimeWarp]=0;
				}
			}
		}

		GLfloat total=remaining[etFireRate]+remaining[etLaserType]+remaining[etSpeedup]+remaining[etShield]+remaining[etTimeWarp];

		if(total==0)
		{
			extraType=etScore;
		}
		else
		{
			p[etFireRate]=remaining[etFireRate]/total;
			p[etLaserType]=remaining[etLaserType]/total;
			p[etSpeedup]=remaining[etSpeedup]/total;
			p[etShield]=remaining[etShield]/total;
			p[etTimeWarp]=remaining[etTimeWarp]/total;

			extraType=(ExtraType)(mysdl::randomProb<int>(extraCount-1,p));
		}

		return(extraType);
	}

	/*
	* See class definition.
	*/
	void Game::destroyedFoe(Player *player,Foe *foe)
	{
		if(player)
		{
			player->addScore(specs.scoreFoe);
			spawnQueued();
		}
	}

	/*
	* See class definition.
	*/
	void Game::destroyedAsteroid(Player *player,Asteroid *asteroid)
	{
		if(player)
		{
			player->addScore(asteroid->type==atSmall ? specs.scoreAsteroidSmall : specs.scoreAsteroidLarge);

			if(asteroid->regular)
			{
				if(killsLeft>0)
				{
					killsLeft--;

					if(killsLeft==0)
					{
						levelComplete=true;
					}
				}
			}

			killsLeftForContainer--;

			if(killsLeftForContainer<=0)
			{
				killsLeftForContainer+=(Uint32)((GLfloat)killsForLevel*0.4);
				game->containerQueue++;
			}


			if(killsLeftForFoe>0)
			{
				killsLeftForFoe--;

				if(killsLeftForFoe<=0)
				{
					game->spawnFoe(0);
				}
			}

			spawnQueued();
		}
	}

	/*
	* See class definition.
	*/
	void Game::spawnQueued()
	{
		if(containerQueue>0)
		{
			spawnContainer();
			containerQueue--;
		}
	}

	/*
	* See class definition.
	*/
	void Game::spawnContainer()
	{
		if(activePlayers>0)
		{
			Container *container=new Container();
			game->spritesAsteroids.addFront(container);
		}
	}

	/*
	* See class definition.
	*/
	void Game::spawnExtra(ExtraType type)
	{
		mysdl::Vector2D pos;

		pos.x=mysdl::random<GLfloat>(specs.playfieldBorderX,resources->vpPlayfield.dimensions.x-specs.playfieldBorderX);
		pos.y=resources->vpPlayfield.dimensions.y+50;

		if(type==etUnknown)
		{
			type=game->getExtraType();
		}

		Extra *extra=new Extra(type,pos);
		game->spritesExtras.addFront(extra);
	}

	/*
	* See class definition.
	*/
	bool Game::spawnFoe(int where)
	{
		if((foes[0])&&(foes[1]))
		{
			return(false);
		}

		if(where==0)
		{
			if((!(foes[0]))&&(!(foes[1])))
			{
				where=(mysdl::random_int(0,1)==0 ? +1 : -1);
			}
			else
			{
				where=(foes[0] ? +1 : -1);
			}
		}

		where=(where>0 ? 1 : 0);

		if(foes[where])
		{
			return(false);
		}

		foes[where]=new Foe(where>0);

		return(true);
	}

	/*
	* See class definition.
	*/
	void Game::setHighscoresChecked(Uint8 checked)
	{
		highscoresChecked=checked;

		while(highscoresChecked<playerCount)
		{
			if(highscorePos[highscoresChecked]>=0)
			{
				state=igsEnterNames;
				((*resources->gui)["pEnterName"])->setVisible(true);
				mysdl::gui::Label *label=((mysdl::gui::Label *)((*resources->gui)["lEnterName"]));
				label->setCaption("Enter your name, player "+mysdl::toStr((int)highscoresChecked+1)+":");
				mysdl::gui::Textfield *textfield=((mysdl::gui::Textfield *)((*resources->gui)["tfEnterName"]));
				textfield->setCaption(previousName[highscoresChecked]);
				textfield->setStyle(highscoresChecked==0 ? resources->lsTextfieldRed : resources->lsTextfieldBlue);
				((*resources->gui)["tfEnterName"])->setFocused(true);
				resources->gui->setActive(true);
				resources->gui->setVisible(true);
				evaluators[0]->enabled=false;

				if(evaluators[1])
				{
					evaluators[1]->enabled=false;
				}

				resources->gui->Evaluator::enabled=true;
				state=igsEnterNames;
				break;
			}
			else
			{
				highscoresChecked++;
			}
		}

		if(highscoresChecked>=playerCount)
		{
			resources->saveHighscores();

			resources->queue->flush();
 			highscoreTime=0;
			state=igsShowHighscores;
			resources->queue->flush();

			if((resources->musicManager)&&(resources->settings.volumeMusic>0))
			{
				if((highscorePos[0]>=0)||((highscorePos[1]>=0)))
				{
					resources->musicManager->addAsLast(resources->musicHighscoresIntro);
					resources->musicManager->addAsLast(resources->musicHighscoresLoop);
					resources->musicManager->skipToNext();
				}
				else
				{
					resources->musicManager->addAsLast(resources->musicGameOverIntro);
					resources->musicManager->addAsLast(resources->musicGameOverLoop);

					if(!resources->musicManager->isPlaying())
					{
						resources->musicManager->play();
					}
				}
			}
		}
	}

	/*
	* See class definition.
	*/
	void Game::spawnAsteroid()
	{
		if(activePlayers)
		{
			GLfloat chanceLarge=(game->type==gtSingleB ? specs.chanceLargeAsteroidB : specs.getChanceLargeAsteroid(level));
			GLfloat chanceEnergy=specs.getChanceAsteroidEnergy(level);
			Uint32 energy=(mysdl::random<GLfloat>(1.0)<chanceEnergy ? mysdl::random_int(specs.getMaxAsteroidEnergy(level)) : 0);

			Asteroid *asteroid=new Asteroid(mysdl::random<GLfloat>(1.0)<chanceLarge ? atLarge : atSmall,energy);

			asteroid->pos.y=resources->vpPlayfield.dimensions.y+asteroid->image->getHeight()+
					(mysdl::random<GLfloat>(resources->vpPlayfield.dimensions.y*0.3));

			asteroid->velocity=specs.getAsteroidVelocity(game->speed);

			GLfloat dy=asteroid->pos.y-specs.asteroidTargetY;
			GLfloat dx=-dy/asteroid->velocity.y*asteroid->velocity.x;

			GLfloat xMin=-dx;
			GLfloat xMax=resources->vpPlayfield.dimensions.x-dx;

			asteroid->pos.x=mysdl::random<GLfloat>(xMin,xMax);

			spritesAsteroids.addFront(asteroid);
			totalAsteroids++;
		}
	}

	/*
	* See class definition.
	*/
	void Game::doPerform(int stage,Uint32 ticks)
	{
		switch(stage)
		{
			case 1:
			{
				if(!paused)
				{
					if(!levelComplete)
					{
						asteroidTimer+=(Uint32)((GLfloat)ticks*timeWarpFactor);

						while(asteroidTimer>asteroidInterval)
						{
							asteroidTimer-=asteroidInterval;
							spawnAsteroid();
						}

						if((type==gtSingleB)&&(players[0]->alive)) // TODO: solve via specs?
						{
							speedupTimer+=(Uint32)((GLfloat)ticks*timeWarpFactor);

							while(speedupTimer>specs.speedupInterval(speed))
							{
								speedupTimer-=specs.speedupInterval(speed);
								players[0]->score+=specs.scoreIncrease;
								speed+=specs.speedIncreaseB;

								if(asteroidInterval>specs.asteroidIntervalMinB)
								{
									asteroidInterval-=specs.asteroidIntervalDecreaseB;

									if(asteroidInterval<specs.asteroidIntervalMinB)
									{
										asteroidInterval=specs.asteroidIntervalMinB;
									}
								}
							}

							extraTimerScore+=(Uint32)((GLfloat)ticks*timeWarpFactor);

							while(extraTimerScore>specs.extraIntervalScore)
							{
								extraTimerScore-=specs.extraIntervalScore;
								spawnExtra(etScore);
							}

							extraTimerSpeedup+=(Uint32)((GLfloat)ticks*timeWarpFactor);

							while(extraTimerSpeedup>specs.extraIntervalExtra)
							{
								extraTimerSpeedup-=specs.extraIntervalExtra;
								spawnExtra(etUnknown);
							}
						}
					}

					for(unsigned int i=0; i<playerCount; i++)
					{
						if(players[i]->alive)
						{
							players[i]->setDirection(mysdl::Vector2D(axisPlayerX[i]->value,-axisPlayerY[i]->value*specs.playerSpeedFactorY));
							players[i]->setShield(triggerPlayerShield[i]->isActive());
						}
					}

					switch(state)
					{
						case igsGame:
						{
							if(levelComplete)
							{
								if(spritesAsteroids.size()==0)
								{
									if(totalAsteroids<=killsForLevel*1.5)
									{
										bonus=specs.scoreBonus;
									}
									else if(totalAsteroids<killsForLevel*3)
									{
										GLfloat factor=((GLfloat)-(totalAsteroids-3*killsForLevel))/((GLfloat)killsForLevel*1.5);
										bonus=Uint32(((GLfloat)specs.scoreBonus)*factor);
									}
									else
									{
										bonus=0;
									}

									haveBonus=(bonus>0);
									pingCount=0;
									fireEnabled=false;
									state=igsBonusFadeIn;
									stateChangeCountdown=3000;
								}
							}

							break;
						}

						case igsBonusFadeIn:
						{
							if(interludeColor.a<1)
							{
								interludeColor.a+=ticks/1000.0;

								if(interludeColor.a>=1)
								{
									interludeColor.a=1;
								}
							}

							stateChangeCountdown-=ticks;

							if(stateChangeCountdown<=0)
							{
								stateChangeCountdown=0;
								state=igsBonusDistribute;
							}

							break;
						}

						case igsBonusDistribute:
						{
							if(bonus>0)
							{
								Uint32 dec=2*(Uint32)ticks;

								pingCount-=ticks;

								if(pingCount<0)
								{
									pingCount+=50; // TODO: move to specs

									if(resources->settings.volumeSfx>0)
									{
										resources->sourcesGeneral->play(resources->samplePing);
									}
								}

								if(bonus<=dec)
								{
									for(Uint32 i=0; i<playerCount; i++)
									{
										players[i]->addScore(bonus);
									}

									bonus=0;
								}
								else
								{
									for(Uint32 i=0; i<playerCount; i++)
									{
										players[i]->addScore(dec);
									}

									bonus-=dec;
								}
							}

							if(bonus==0)
							{
								state=igsBonusPause;
								stateChangeCountdown=3000;
							}

							break;
						}

						case igsBonusPause:
						{
							stateChangeCountdown-=ticks;

							if(stateChangeCountdown<=0)
							{
								stateChangeCountdown=1500;
								state=igsBonusFadeOut;
							}

							break;
						}

						case igsBonusFadeOut:
						{
							if(interludeColor.a>0)
							{
								interludeColor.a-=ticks/1000.0;

								if(interludeColor.a<=0)
								{
									interludeColor.a=0;
								}
							}

							stateChangeCountdown-=ticks;

							if(stateChangeCountdown<=0)
							{
								stateChangeCountdown=0;
								state=igsGame;
								initLevel(level+1);
							}

							break;
						}

						case igsShowHighscores:
						{
							if(highscoreTime<specs.restartDelay)
							{
								highscoreTime+=ticks;
							}
						}

						default:
						{
							// nothing
						}
					}

					Uint32 intTicks=(Uint32)ticks;

					if(timeWarpDuration>0)
					{
						if(timeWarpDuration>intTicks)
						{
							timeWarpDuration-=intTicks;
						}
						else
						{
							timeWarpDuration=0;
							timeWarpFactorDestination=1;
						}
					}

					if(timeWarpFactor!=timeWarpFactorDestination)
					{
						if(timeWarpDuration>0)
						{
							timeWarpFactor-=specs.timeWarpStartSpeed*ticks;

							if(timeWarpFactor<timeWarpFactorDestination)
							{
								timeWarpFactor=timeWarpFactorDestination;
							}
						}
						else
						{
							timeWarpFactor+=specs.timeWarpStopSpeed*ticks;

							if(timeWarpFactor>timeWarpFactorDestination)
							{
								timeWarpFactor=timeWarpFactorDestination;
							}
						}
					}

					spritesPlayer.perform(ticks*timeWarpFactor);
					spritesAsteroids.perform(ticks*timeWarpFactor);
					spritesExtras.perform(ticks*timeWarpFactor);
					spritesFoes.perform(ticks*timeWarpFactor);
					spritesFoeLaser.perform(ticks*timeWarpFactor);

					spritesLaser.perform(ticks*timeWarpFactor/2.0);
					spritesLaser.testCollisions(&spritesFoes);
					spritesLaser.perform(ticks*timeWarpFactor/2.0);

					spritesPlayer.testCollisions(&spritesAsteroids);
					spritesPlayer.testCollisions(&spritesExtras);
					spritesPlayer.testCollisions(&spritesFoes);
					spritesLaser.testCollisions(&spritesAsteroids);
					spritesLaser.testCollisions(&spritesFoes);
					spritesFoeLaser.testCollisions(&spritesAsteroids);
					spritesFoeLaser.testCollisions(&spritesPlayer);

					if(state<igsEnterNames)
					{
						resources->starfield->speedFactor=speed*specs.starSpeedFactor*timeWarpFactor;
					}
					else
					{
						if(resources->starfield->speedFactor<specs.starfieldMenuSpeed)
						{
							resources->starfield->speedFactor+=specs.starfieldMenuSpeedStep*ticks;

							if(resources->starfield->speedFactor>specs.starfieldMenuSpeed)
							{
								resources->starfield->speedFactor=specs.starfieldMenuSpeed;
							}
						}
						else if(resources->starfield->speedFactor>specs.starfieldMenuSpeed)
						{
							resources->starfield->speedFactor-=specs.starfieldMenuSpeedStep*ticks;

							if(resources->starfield->speedFactor<specs.starfieldMenuSpeed)
							{
								resources->starfield->speedFactor=specs.starfieldMenuSpeed;
							}
						}
					}

					for(unsigned int i=0; i<playerCount; i++)
					{
						if(players[i]->alive)
						{
							if(players[i]->hud.shieldEnergy<players[i]->shieldEnergy)
							{
								players[i]->hud.shieldEnergy+=ticks*5.0;

								if(players[i]->hud.shieldEnergy>players[i]->shieldEnergy)
								{
									players[i]->hud.shieldEnergy=players[i]->shieldEnergy;
								}
							}
							else
							{
								players[i]->hud.shieldEnergy=players[i]->shieldEnergy;
							}

							if(players[i]->hud.switchesLeft>0)
							{
								players[i]->hud.switchCountdown-=ticks;

								while((players[i]->hud.switchesLeft>0)&&(players[i]->hud.switchCountdown<=0))
								{
									players[i]->hud.actual=!players[i]->hud.actual;
									players[i]->hud.switchesLeft--;

									if(players[i]->hud.switchesLeft>0)
									{
										players[i]->hud.switchCountdown+=specs.hudSwitchTime;
									}
								}
							}
						}
					}

					if((!over)&&(activePlayers==0))
					{
						gameOver();
					}

					if((over)&&(leaving))
					{
						terminate=true;
					}
				}
			}
		}

		GameState::doPerform(stage,ticks);
	}

	/*
	* See class definition.
	*/
	void Game::drawHud(mysdl::Vector2D pos,int speed,int fireType,int fireRate,int timewarp,GLfloat shield)
	{
		specs.colorHudBack.apply();
		resources->imgHudFull->draw(pos);

		specs.colorHudFront.apply();

		if(shield>0)
		{
			int height=(int)(((GLfloat)(resources->imgHudShield->getHeight()))*shield);

			resources->imgHudShield->draw(mysdl::Rect(pos.x-resources->imgHudShield->getHotspot().x,
																								pos.y+resources->imgHudShield->getHeight()+resources->imgHudShield->getHotspot().y-height,
																								resources->imgHudShield->getWidth(),height));
		}

		if(speed>0)
		{
			resources->imgHudSpeed[speed-1]->draw(pos);
		}

		resources->imgHudFire[fireType][fireRate]->draw(pos);

		if(timewarp>0)
		{
			resources->imgHudTimewarp[timewarp-1]->draw(pos);
		}
	}

	/*
	* See class definition.
	*/
	void Game::doDraw(int stage)
	{
		switch(stage)
		{
			case 1:
			{
				spritesExtras.draw();
				spritesLaser.draw();
				spritesFoeLaser.draw();
				spritesPlayer.draw();
				spritesFoes.draw();
				spritesAsteroids.draw();

				resources->contextHud.apply();

				red.apply();
				mysdl::resetCursor();
				mysdl::moveCursor(mysdl::Vector2D(770,450));
				resources->fontStandard->printf(mysdl::jhRight,"%5d",players[0]->score);

				if(players[0]->alive)
				{
					mysdl::resetCursor();
					white.apply();

					if(players[0]->hud.actual)
					{
						drawHud(specs.hudPosPlayer1,players[0]->speedupCount,players[0]->laserLevel,players[0]->fireRateCount,players[0]->timeWarpLevel,players[0]->hud.shieldEnergy/specs.shieldEnergyMax);
					}
					else
					{
						drawHud(specs.hudPosPlayer2,players[0]->hud.speedup,players[0]->hud.fireType,players[0]->hud.fireRate,players[0]->hud.timewarp,players[0]->hud.shieldEnergy/specs.shieldEnergyMax);
					}
				}

				if(players[1])
				{
					blue.apply();

					mysdl::resetCursor();
					mysdl::moveCursor(mysdl::Vector2D(30,450));
					resources->fontStandard->printf(mysdl::jhLeft,"%5d",players[1]->score);
				}

				if((players[1])&&(players[1]->alive))
				{
					mysdl::resetCursor();
					white.apply();

					if(players[1]->hud.actual)
					{
						drawHud(mysdl::Vector2D(60,540),players[1]->speedupCount,players[1]->laserLevel,players[1]->fireRateCount,players[1]->timeWarpLevel,players[1]->hud.shieldEnergy/specs.shieldEnergyMax);
					}
					else
					{
						drawHud(mysdl::Vector2D(60,540),players[1]->hud.speedup,players[1]->hud.fireType,players[1]->hud.fireRate,players[1]->hud.timewarp,players[1]->hud.shieldEnergy/specs.shieldEnergyMax);
					}
				}

				switch(state)
				{
					case igsBonusFadeIn:
					case igsBonusDistribute:
					case igsBonusPause:
					case igsBonusFadeOut:
					{
						interludeColor.apply();
						mysdl::resetCursor();
						mysdl::moveCursor(mysdl::Vector2D(400,320));
						resources->fontStandard->printf(mysdl::jhCenter,"Stage %d complete.",level);

						if(haveBonus)
						{
							mysdl::resetCursor();
							mysdl::moveCursor(mysdl::Vector2D(400,280));
							resources->fontStandard->printf(mysdl::jhCenter,"Bonus: %d",(int)bonus);
						}

						break;
					}

					case igsShowHighscores:
					{
						mysdl::Color(1.0,0.3,0.3,1.0).apply();
						mysdl::resetCursor();
						mysdl::moveCursor(mysdl::Vector2D(400,530));
						resources->fontStandard->printf(mysdl::jhCenter,"G A M E   O V E R");

						mysdl::Color(1.0,1.0,1.0,1.0).apply();
						mysdl::resetCursor();
						mysdl::moveCursor(mysdl::Vector2D(400,70));
						resources->fontStandard->printf(mysdl::jhCenter,"Press fire to start a new game.");

						displayHighscores(400,300,getHighscores(),highscorePos[0],highscorePos[1]);
						break;
					}

					default:
					{
						// nothing
					}
				}
			}

			default:
			{
				GameState::doDraw(stage);
			}
		}
	}

	/*
	* See class definition.
	*/
	void Game::initLevel(Uint32 level)
	{
		this->level=level;
		levelComplete=false;

		killsForLevel=specs.getKillsForLevel(level);
		killsLeft=killsForLevel+(type==gtMulti ? killsForLevel : 0);
		killsLeftForContainer=(Uint32)((GLfloat)killsForLevel*0.4);
		killsLeftForFoe=level<specs.foeFirstLevel ? 0 : killsForLevel/2;

		spawnQueued();

		if(type!=gtSingleB)
		{
			speed=specs.getSpeedForLevel(game->level);
			asteroidInterval=specs.getAsteroidInterval(level);
		}
		else
		{
			speed=specs.startSpeedModeB;
			asteroidInterval=specs.startAsteroidIntervalModeB;
		}

		asteroidCount=0; // TODO: what's this?

		fireEnabled=(type!=gtSingleB);
		state=igsGame;
		totalAsteroids=0;
	}

	/*
	* See class definition.
	*/
	void Game::gameOver()
	{
		over=true;

		if(terminated)
		{
			setHighscoresChecked(playerCount);
		}
		else
		{
			for(Uint32 i=0; i<playerCount; i++)
			{
				highscorePos[i]=game->getHighscores()->getPos(HighscoreEntry("dummy",players[i]->score));

				if(i==1)
				{
					if((players[1]->score>players[0]->score)&&(highscorePos[0]>=0))
					{
						highscorePos[0]++;

						if(highscorePos[0]>=10)
						{
							highscorePos[0]=-1;
						}
					}
					else if((players[1]->score<players[0]->score)&&(highscorePos[1]>=0))
					{
						highscorePos[1]++;

						if(highscorePos[1]>=10)
						{
							highscorePos[1]=-1;
						}
					}
				}
			}

			setHighscoresChecked(0);
		}

		resources->queue->flush();
	}

	/*
	* See class definition.
	*/
	HighscoreTable *Game::getHighscores()
	{
		switch(type)
		{
			case gtSingleA:
				return(&(resources->highscores.singleA));

			case gtSingleB:
				return(&(resources->highscores.singleB));

			default:
				return(&(resources->highscores.multi));
		}
	}

	/*
	* See class definition.
	*/
	void Game::startMusic()
	{
		if((resources->musicManager)&&(resources->currentVolumeMusic>0))
		{
			resources->musicManager->clearQueue();
			resources->musicManager->addAsLast(type==gtSingleB ? resources->musicGameBIntro : resources->musicGameAIntro);
			resources->musicManager->addAsLast(type==gtSingleB ? resources->musicGameBLoop : resources->musicGameALoop);

			if(resources->musicManager->isPlaying())
			{
				resources->musicManager->skipToNext();
			}
			else
			{
				resources->musicManager->play();
			}
		}
	}
};
