/**
 * Max Fighter
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://maxfighter.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
#ifndef MAXFIGHTER_GAME_H
#define MAXFIGHTER_GAME_H

#include <mysdl/mysdl.h>
#include <mysdl/sprites/sprites.h>

#include "resources.h"
#include "specs.h"

namespace maxfighter
{
	class Player;
	class Extra;

	/** Current state of the Game state machine. */
	enum InGameState {igsGameIntro=0,igsGame,igsBonusFadeIn,igsBonusDistribute,igsBonusPause,igsBonusFadeOut,igsEnterNames,igsShowHighscores};

	/** Number of available extra types, as described by the ExtraType enum. */
	static const unsigned int extraCount=6;

	/** Available extra types. */
	enum ExtraType {etUnknown=-1,etLaserType=0,etSpeedup=1,etFireRate=2,etShield,etTimeWarp,etScore};

	/** Available game modes. */
	enum GameType {gtSingleA,gtSingleB,gtMulti};

	/** Available asteroid types. */
	enum AsteroidType {atSmall,atLarge};

	/** Available player types, depending on the game mode. */
	enum PlayerType {ptSingle,ptLeft,ptRight};

	/**
	 * Base class for sprites in Max Fighter. These are ImageSprites which
	 * supported delayed deletion. Calling deleteMe() will mark the sprite
	 * for deletion but keep it until the next garbage collection, so it is
	 * still available for the current game loop.
	 */
	class MaxfighterSprite : public mysdl::sprites::ImageSprite
	{
		protected:
			/** Whether the sprite has already been deleted (i.e. marked for
			 *  delayed deletion. */
			bool deleted;

		public:
			/**
			 * Constructor.
			 */
			MaxfighterSprite();

			/**
			 * Destructor.
			 */
			virtual ~MaxfighterSprite();

			/**
			 * Deletes the sprite, or, more accurately, marks it for delayed
			 * deletion. The sprite instance will be deleted during the next
			 * call to TODO: ???
			 */
			void deleteMe();
	};

	/**
	 * A sprite showing an explosion. The sprite will automatically delete
	 * itself when the explosion animation has finished.
	 */
	class ExplosionSprite : public MaxfighterSprite
	{
		private:
			/** Fixed explosions don't move/scroll with the starfield.
			 *  TODO: we should be able to delete this and simply set velocity() accordingly */
			bool fixed;

		public:
			/** Movement vector. */
			mysdl::Vector2D velocity;

			/** Current animation reference. When this has finished, the sprite
			 *  will be deleted. */
			mysdl::gfx::AnimationReference *anim;

			/**
			 * Constructor.
			 */
			ExplosionSprite(mysdl::gfx::Animation *animation,bool fixed=false);

			/**
			 * Moves the sprite and updates the animation. Marks the sprite for
			 * deletion when the animation has finished. */
			void perform(GLfloat ticks);
	};

	/**
	 * The shield for a player sprite. This will be displayed above the player
	 * to which it belongs. */
	class Shield : public mysdl::sprites::ImageSprite
	{
		public:
			/** Player to which this shield belongs. */
			Player *player;

			/** Whether the shield is currently collapsing. TODO: this seems unused? */
			bool collapsing;

			/** Whether the player has released the shield. TODO: ? */
			bool released;

			/**
			 * Constructor.
			 *
			 * @param player Player to which the shield belongs.
			 */
			Shield(Player *player);

			/**
			 * Destroys most objects when they collide with an active shield.
			 */
			void onCollision(mysdl::sprites::Sprite *sprite);
	};

	/**
	 * Snapshot of a player's state as displayed in a HUD. This is stored
	 * separately so that new states can be shown by switching between the
	 * current and previous state for a few seconds. In this case, the
	 * HudState instance is used to store the previous state, and is
	 * updated once the transition effect has finished.
	 */
	class HudState
	{
		public:
			/** Number of speedup extras collected. */
			int speedup;

			/** Number of fire rate extras collected. */
			int fireRate;

			/** Number of fire type extras collected. */
			int fireType;

			/** Number of timewarp items collected. */
			int timewarp;

			/** Shield energy (0=none, 1=full). */
			GLfloat shieldEnergy;

			/** Number of milliseconds until the next switch. */
			GLfloat switchCountdown;

			/** Number of remaining switches until the HUD state instance
			 *  is updated. */
			int switchesLeft;

			/** Whether the HUD currently displays the current state. If
			 *  this is false, the state as recorded in the HudState
			 *  instance is displayed. */
			bool actual;
	};

	/**
	 * A human-controlled player.
	 */
	class Player : public MaxfighterSprite
	{
		public:
			/** Whether the player is still alive. */
			bool alive;

			/** Sprite for showing the engine animation. */
			mysdl::sprites::ImageSprite *engine;

			/** Shield sprite. */
			Shield *shield;

			/** Movement vector. */
			mysdl::Vector2D direction;

			/** Current score for this player. */
			Uint32 score;

			/** When the player's score reaches or surpasses this value, a
			 *  new container will be added to the game's queue. */
			Uint32 nextContainer;

			/** The minimum delay between two laser shots. */
			GLfloat laserDelay;

			/** The speed factor for movement vectors of laser shots fired by
			 *  this player. */
			GLfloat laserSpeed;

			/** The laser upgrade level. */
			short int laserLevel;

			/** The speed factor for player movement. Every tick, the player
			 *  moves by the movement vector multiplied by the speed. */
			GLfloat speed;

			/** Number of collected speedup extras. */
			int speedupCount;

			/** NUmber of collected fire rate extras. */
			int fireRateCount;

			/** Number of milliseconds until the primary/secondary laser may
			 *  be fired again. */
			GLfloat laserTimer,laserTimerSecondary;

			/** Minimum/maximum X positions. */
			GLfloat xMin,xMax;

			/** Reference to the current player animation. */
			mysdl::gfx::AnimationReference *anim;

			/** Remaining shield energy. */
			GLfloat shieldEnergy;

			/** Number of collected timewarp extras. */
			int timeWarpLevel;

			/** Whether the shield is currently active. */
			bool shielding;

			/** Number of milliseconds remaining in timewarp mode. */
			Uint32 timeWarpDuration;

			/** Whether the engine is currently in "stutter" mode, which uses
			 *  the special stutter animation. */
			bool engineStutter;

			/** Stored of the HUD. This may differ from the current extra
			 *  state if the HUD is currently in "flashing mode". */
			HudState hud;

			/**
			 * Constructor.
			 *
			 * @param type Player type (single, left or right).
			 */
			Player(PlayerType type);

			/**
			 * Destructor.
			 */
			~Player();

			/**
			 * When the player is hit by another object while not shileding,
			 * he looses extras (twice as much if the other object was a
			 * foe). When the player didn't have any extras, he explodes.
			 */
			void hit(MaxfighterSprite *sprite);

			/**
			 * When the player collides with an extra, he collects it. When
			 * he collides with any other object while not shieldinghit() is
			 * called.
			 */
			void onCollision(mysdl::sprites::Sprite *sprite);

			/**
			 * Updates a player's shield, engine and timewarp sprites and
			 * settings as well as his position and laser state. Also makes
			 * sure the position stays in the allowed range, and that two
			 * players don't get too close to each other.
			 */
			void perform(GLfloat ticks);

			/**
			 * Fires the primary and secondary laser if possible.
			 */
			void fire();

			/**
			 * Activates or deactivates the shield.
			 */
			void setShield(bool shield);

			/**
			 * Sets the current movement vector.
			 */
			void setDirection(mysdl::Vector2D(direction));

			/**
			 * Adds points to the player's score. May add a new container to
			 * the game's queue.
			 */
			void addScore(Uint32 add);

			/**
			 * Displays an explosion for the player.
			 */
			void explode();

			/**
			 * Initiates timewarp for this player if possible. Reduces the
			 * number of available timewarp items.
			 */
			void startTimeWarp();

			/**
			 * Collects the specified extra if it hasn't already been
			 * collected and updates player stats.
			 *
			 * @return true iff the extra could be collected.
			 */
			bool collectExtra(Extra *extra);

			/**
			 * Sets the number of collected speedup extras and updates the
			 * player's stats accordingly.
			 */
			void setSpeedupCount(int count);

			/**
			 * Sets the number of collected fire rate extras and updates the
			 * player's stats accordingly.
			 */
			void setFireRateCount(int count);

			/**
			 * Updates the HUD to represent the current player stats.
			 *
			 * @param startTimer If this is true, the HUD will be set to
			 *                   "flash mode".
			 */
			void updateHud(bool startTimer);
	};

	/**
	 * A laser shot fired by a player or foe.
	 */
	class Laser : public MaxfighterSprite
	{
		public:
			/** The player or foe who fired this shot. */
			MaxfighterSprite *source;

			/** Movement vector. */
			mysdl::Vector2D velocity;

			/** Remaining milliseconds until the sprite is deleted. -1 means
			 *  the sprite will not be deleted until it leaves the screen. */
			GLfloat ttl;

			/** Speed used to fade out the shot. If this is 0, the shot will
			 *  not fade. */
			GLfloat fadeSpeed;

			/**
			 * Constructor.
			 *
			 * @param img       Laser image
			 * @param source    The player or foe who fired the shot
			 * @param ttl       Remaining milliseconds until the sprite is deleted.
			 * @param fadeSpeed Speed used to fade out the shot.
			 */
			Laser(mysdl::gfx::Image *img,MaxfighterSprite *source,GLfloat ttl=-1,GLfloat fadeSpeed=0);

			/**
			 * Moves the shot and fades it. Deletes the sprite once ttl reaches 0 or
			 * when the shot has left the screen.
			 */
			void perform(GLfloat ticks);
	};

	/**
	 * An asteroid (small or large).
	 */
	class Asteroid : public MaxfighterSprite
	{
		public:
			/** Asteroid type (small or large). */
			AsteroidType type;

			/** Remaining energy. An asteroid is destroyed by a laser shot only when
			 * its energy is 0, otherwise the hit simply decreases the asteroid's
			 * energy. */
			int energy;

			/** Only regular asteroids are registered as kills necessary for
			 *  advancing to the next level. */
			bool regular;

			/** Movement vector. */
			mysdl::Vector2D velocity;

			/** Whether the asteroid should be destroyed at the next update. This
			 *  is used to delay the call that triggers delayed deletion, so that
			 *  the asteroid does still trigger collisions until then. */
			bool destroy;

			/**
			 * Constructor.
			 *
			 * @param type    Asteroid type (small or large)
			 * @param energy  Asteroid energy.
			 * @param regular Whether the asteroid is a regular asteroid.
			 */
			Asteroid(AsteroidType type=atSmall,int energy=0,bool regular=true);

			/**
			 * Destructor.
			 */
			virtual ~Asteroid();

			/**
			 * Triggers an explosion for the asteroid and sets the "destroyed"
			 * flag.
			 *
			 * @param complete Whether the asteroid should be completely destroyed.
			 *                 This makes a different only for large asteroids,
			 *                 which are split into two smaller asteroids if this
			 *                 is false.
			 */
			void explode(bool complete);

			/**
			 * Updates the asteroid's position or deletes the asteroid if it
			 * is marked as destroyed or has left the screen. */
			void perform(GLfloat ticks);

			/**
			 * When the asteroid collides with a laser shot, its energy is
			 * reduced. When the energy was 0, the asteroid is destroyed.
			 */
			void onCollision(mysdl::sprites::Sprite *sprite);
	};

	/**
	 * A container carrying an extra item.
	 */
	class Container : public MaxfighterSprite
	{
		public:
			/** Type of the extra carried by the container. The animation
			 *  for the container also depends on this. */
			ExtraType extraType;

			/**
			 * Constructor.
			 */
			Container();

			/**
			 * Creates an explosion for the container and releases the
			 * extra stored inside, then marks the container for deletion.
			 */
			void explode();

			/**
			 * Moves the container. Deletes it when it leaves the screen.
			 */
			void perform(GLfloat ticks);

			/**
			 * When the container is hit by a laser shot, it explodes.
			 * TODO: why is this separate from other collisions?
			 */
			void onCollision(mysdl::sprites::Sprite *sprite);
	};

	/**
	 * An extra which can be collected by a player.
	 */
	class Extra : public MaxfighterSprite
	{
		public:
			/** Type of the extra. */
			ExtraType extraType;

			/** Whether this extra has already been collected. Prevents
			 *  collecting the same extra multiple times while it is
			 *  fading out. */
			bool collected;

			/** Movement vector. May include a horizontal component when
			 *  the extra is "ejecting" from a hit player. */
			mysdl::Vector2D velocity;

			/** Number of milleseconds until the extra is deleted. If this
			 *  is 0 (or smaller), the extra is only deleted once it leaves
			 *  the screen. */
			Uint32 ticks;

			/**
			 * Constructor.
			 *
			 * @param extraType Type of this extra. etUnknown is not allowed.
			 * @param pos       Starting position.
			 */
			Extra(ExtraType extraType,mysdl::Vector2D pos);

			/**
			 * Updates the vector's position and tick count. Fades out the
			 * extra if necessary, and deletes it when its tick count
			 * reaches 0 or it leaves the screen.
			 */
			void perform(GLfloat ticks);
	};

	/**
	 * A "pirate", appearing either on the left or the right side
	 * of the screen. The foe stays on the screen until it is
	 * destroyed or the game is over.
	 */
	class Foe : public MaxfighterSprite
	{
		private:
			/** Number of milliseconds before the next laser volley may be
			 *  fired. */
			Uint32 fireDelay;

			/** Number of shots per volley. */
			Uint32 volleyCount;

			/** Number of milliseconds between shots in a volley. */
			Uint32 volleyDelay;

			/** Amount by which the velocity may change each millisecond. */
			GLfloat acceleration;

			/** Determines how quickly the speed will change depending on
			 *  the distance to the targeted player. */
			float distanceSpeedFactor;

			/** Number of milliseconds left in "flooring mode". Flooring is
			 *  the maneuver where the foe stays at the lower edge of the
			 *  screen regardless of player positions. */
			Uint32 floorCountdown;

			/** Whether the foe is currently in kamikaze mode, i.e. moving
			 *  downwards into the player instead of keeping his distance. */
			bool kamikaze;

			/** Whether the foe is on the right side of the screen. If this
			 *  is false, the foe is on the left side. */
			bool right;

			/** Current movement vector. The x component is always 0. */
			GLfloat velocity;

			/** Reference to the current animation. */
			mysdl::gfx::AnimationReference *anim;

			/** Number of shots left in the current volley. */
			Uint32 volley;

			/** Number of milliseconds until the next shot may be fired. */
			GLfloat laserTimer;

			/** Whether the foe is still alive. */
			bool alive;

			/** Hull integrity. If this is positive, laser hits will reduce
			 *  this value. If it is 0 and the foe is hit, he explodes. */
			int hull;

			/** Whether the foe is leaving the screen (because all players
			 *  have died). */
			bool leaving;

		public:
			/**
			 * Constructor.
			 *
			 * @param right Whether the foe appears on the right side of the
			 *              screen. If this is true, he appears on the left
			 *              side.
			 */
			Foe(bool right);

			/**
			 * Destructor.
			 */
			~Foe();

			/**
			 * Performs the foe's very simple "thinking", updates his position
			 * and tries to fire the laser.
			 */
			void perform(GLfloat ticks);

			/**
			 * Fires the foe's laser, if it is currently charged.
			 */
			void fire();

			/**
			 * When the foe collides with a player's laser shot, its hull is
			 * decreased. If the hull was 0, the foe explodes.
			 */
			void onCollision(mysdl::sprites::Sprite *sprite);

			/**
			 * Displays an explosion and deletes the foe sprite.
			 */
			void explode();
	};

	/**
	 * The Game state, representing an active game and containing all the
	 * necessary sprites. The Game state is itself a state machine. The
	 * "state" variable determines the current state.
	 */
	class Game : public GameState
	{
		public:
			/** Input evaluators for the available players. This allows using
			 *  different evaluators for the first player in single and
			 *  multiplayer modes. */
			mysdl::InputEvaluator *evaluators[2];

			/** Triggers for the player movement axes. */
			mysdl::Trigger *axisPlayerX[2],*axisPlayerY[2];

			/** Triggers for the player's shields. */
			mysdl::Trigger *triggerPlayerShield[2];

			/** Game type (Single A, Single B or Multiplayer). */
			GameType type;

			/** true if the game is over. */
			bool over;

			/** Whether the player is leaving the game. TODO: This might
			 *  actually happen immediately by now, in which case this
			 *  variable could be eliminated. */
			bool leaving;

			/** Whether the current level has been completed. If this is
			 *  true, no new asteroids are spawned, and a state change to
			 *  the beginning of the "interlude" is imminent. */
			bool levelComplete;

			/** Current game speed. */
			GLfloat speed;

			/** Remaining number of asteroids that have to be destroyed
			 *  before the level is completed. */
			Uint32 killsLeft;

			/** Current level, 1 being the first. */
			int level;

			/** Number of players (living or dead) in this game. */
			Uint32 playerCount;

			/** Number of players still alive. When this reaches 0, the game
			 *  is over. */
			Uint32 activePlayers;

			/** Number of milliseconds between spawning new asteroids. */
			Uint32 asteroidInterval;

			/** Number of milliseconds since the last asteroid was
			 *  spawned. */
			Uint32 asteroidTimer;

			/** Number of milliseconds until the next game speedup in
			 *  game mode B. */
			Uint32 speedupTimer;

			/** Number of milliseconds until the next score extra is spawned
			 *  in game mode B. */
			Uint32 extraTimerScore;

			/** Number of milliseconds until the next speedup extra is
			 *  spawned in game mode B. */
			Uint32 extraTimerSpeedup;

			/** Sprites for players, engines and shields. */
			mysdl::sprites::SpriteEngine spritesPlayer;

			/** Laser sprites fired by players. */
			mysdl::sprites::SpriteEngine spritesLaser;

			/** Sprites for asteroids and containers. */
			mysdl::sprites::SpriteEngine spritesAsteroids;

			/** Extra sprites. */
			mysdl::sprites::SpriteEngine spritesExtras;

			/** Sprites for foes. */
			mysdl::sprites::SpriteEngine spritesFoes;

			/**  Laser sprites fired by foes. */
			mysdl::sprites::SpriteEngine spritesFoeLaser;

			/** The two player sprites. In single player games, players[1]
			 *  is 0. */
			Player *players[2];

			/** The left (0) and right (1) foe. These are 0 when there
			 *  currently is no foe on that side. */
			Foe *foes[2];

			/** Current state of the game state machine. */
			InGameState state;

			/** Number of milliseconds before switching to the next state.
			 *  Only some states switch automatically. */
			GLfloat stateChangeCountdown;

			/** Remaining bonus. This is initialised when switching to the
			 *  first intermission state and then counts down to 0 during
			 *  the intermission. */
			Uint32 bonus;

			/** Number of regular asteroids currently in the game. */
			Uint32 asteroidCount;

			/** Current colour of the font for the interlude text. Used to
			 *  fade that text in and out. */
			mysdl::Color interludeColor;

			/** Number of players already "checked" for highscore entry. In
			 *  fact, player's scores are checked in advance, and then this
			 *  variable counts through all players, halting every time the
			 *  referenced player has a highscore and waiting until the
			 *  player's name has been entered. */
			Uint8 highscoresChecked;

			/** Whether players are currently able to fire. Used to disable
			 *  player fire during the intermission states. */
			bool fireEnabled;
			Uint32 timeWarpDuration;
			GLfloat timeWarpFactor;
			GLfloat timeWarpFactorDestination;

			/** Number of containers in the queue. At the beginning of each
			 *  level, and whenever an asteroid is destroyed, the queue is
			 *  checked and one container released if necessary. */
			Uint32 containerQueue;

			/** Number of milliseconds until the next "ping" sound is played
			 *  while counting down and distributing the bonus. */
			int pingCount;

			/** Number of milliseconds the highscores have already been
			 *  displayed, to prevent the players from accidentally starting
			 *  a new game right away. This may stop counting once the
			 *  threshhold that allows starting a new game is reached. */
			int highscoreTime;

			/** Whether there is/was any bonus for the finished level.
			 *  Initialised when calculating the bonus points and used to
			 *  decide whether the bonus text is displayed. */
			bool haveBonus;

			/** Number of asteroids that have to be destroyed before the next
			 *  extra container is spawned. */
			int killsLeftForContainer;

			/** Number of asteroids that have to be destroyed before the next
			 *  foe is spawned. */
			int killsLeftForFoe;

			/** Total number of asteroids spawned during this level. */
			int totalAsteroids;

			/** Total number of asteroids that have to be destroyed to finish
			 *  the current level. */
			int killsForLevel;

			/** Highscore position indexes for both players. -1 means that a
			 *  player did not have a highscore. Used to highlight the
			 *  player's entries when displaying the highscores. */
			int highscorePos[2];

			/** Whether the game was manually terminated. In this case,
			 *  player's scores are not checked against the highscore table
			 *  during the game over section. */
			bool terminated;

			/**
			 * Constructor.
			 *
			 * @param type Single A, Single B or Multiplayer.
			 */
			Game(GameType type);

			/**
			 * Destructor.
			 */
			~Game();

			/**
			 * Initialises a game. Must be called from the constructor. Also
			 * called by restart() to initialise a new game.
			 */
			void init();

			/**
			 * Cleans up game data after a game has terminated.
			 */
			void uninit();

			/**
			 * Starts a new game by uninitialising the old data and
			 * initalisng the new game (using uninit() and init()) without
			 * destroying the existing Game instance.
			 */
			void restart();

			/**
			 * Called when a player destroyed an asteroid. Updates the number
			 * of destroyed asteroids, awards points to the player and
			 * updates the container and foe queues if necessary, then spawns
			 * any queued containers/foes.
			 *
			 * @param player   The player that destroyed the asteroid.
			 * @param asteroid The asteroid that was destroyed.
			 */
			void destroyedAsteroid(Player *player,Asteroid *asteroid);

			/**
			 * Called when a player destroyed a foe. Awards points to the
			 * player and spawns
			 * any queued containers/foes.
			 *
			 * @param player The player that destroyed the foe.
			 * @param foe    The foe that was destroyed.
			 */
			void destroyedFoe(Player *player,Foe *foe);

			/**
			 * Initialises game resources when the Game state is activated.
			 */
			void activate();

			/**
			 * Uninitialises game resources when the Game state is
			 * deactivated.
			 */
			void deactivate();

			/**
			 * Handles input events, including player controls.
			 */
			void handleEvent(const mysdl::Trigger *e);

			/**
			 * Performs game state updates, i.e. everything that happens once
			 * per game loop (input checks, sprite movements, collision
			 * tests, state changes etc.)
			 *
			 * @param stage Performance stage. Game state is only updated
			 *              during stage 1.
			 * @param ticks Milliseconds since last perform.
			 */
			void doPerform(int stage,Uint32 ticks);

			/**
			 * Draws the current game state.
			 *
			 * @param stage Drawing stage. Game state is only drawn during
			 *              stage 1.
			 */
			void doDraw(int stage);

			/**
			 * Spawns a new asteroid if there are any players left. The type,
			 * speed and energy of the new asteroid depend on the current
			 * level.
			 */
			void spawnAsteroid();

			/**
			 * Spawns a new container if there are any players left. The
			 * container is initialised with an unknown extra, which means
			 * that getExtraType() will be called to determine the type
			 * of extra stored in the container.
			 */
			void spawnContainer();

			/**
			 * Spawns a new foe. If there are already two active foes,
			 * nothing happens.
			 *
			 * @param where The side of the screen on which the foe should be
			 *              spawned. This is only a preference; if there
			 *              already is a foe on the specified side, the new
			 *              foe will be spawned on the other side.
			 */
			bool spawnFoe(int where);

			/**
			 * If the container queue is not empty, spawn one container.
			 */
			void spawnQueued();

			/**
			 * Spawns an extra of the specified type.
			 *
			 * @param type Extra type. If this is etUnknown, getExtraType()
			 *             is called to determine the extra type.
			 */
			void spawnExtra(ExtraType type);

			/**
			 * Updates the game's settings to match the specified level.
			 *
			 * @param level The level to which the game should be set. 1 is
			 *              the first level.
			 */
			void initLevel(Uint32 level);

			/**
			 * Ends the game. If the game was not manually termined, the
			 * player's scores are checked against the highscore table, and
			 * they may be prompted to enter their names.
			 */
			void gameOver();

			/**
			 * Gets the highscore table for the current game type.
			 */
			HighscoreTable *getHighscores();

			/**
			 * Gets the type for the next extra, depending on the current
			 * level and the current state of upgrades of the two players.
			 */
			ExtraType getExtraType();

			/**
			 * Draws a HUD ath the specified position.
			 *
			 * @param pos      Position (relative to current position).
			 * @param speed    Number of speedups collected.
			 * @param fireType Number of fire type upgrades collected.
			 * @param fireRate Number of fire rate upgrades collected.
			 * @param timewarp Number of timewarp items collected.
			 * @param shield   Shield energy (0=none, 1=full).
			 */
			void drawHud(mysdl::Vector2D pos,int speed,int fireType,int fireRate,int timewarp,GLfloat shield);

			/**
			 * Checks the next player's highscore position. If the player has
			 * a place in the highscore table, display a prompt to enter the
			 * player's name, otherwise continue until a highscore player is
			 * found or all players have been checked. If all players have
			 * been checked, change the game state to highscore display mode.
			 */
			void setHighscoresChecked(Uint8 highscoresChecked);

			/**
			 * Starts the ingame music if music playback is enabled in the
			 * settings.
			 */
			void startMusic();

			/**
			 * Switches back to the ingame menu when the sound settings menu
			 * is closed.
			 */
			void closeWindowSoundSettings(bool confirm);

			/**
			 * Switches back to the ingame menu when a controls menu is
			 * closed.
			 */
			void closeWindowControls(bool confirm);
	};
};

#endif
