/**
 * Max Fighter
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://maxfighter.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
#ifndef MAXFIGHTER_RESOURCES_H
#define MAXFIGHTER_RESOURCES_H

#include <mysdl/controls.h>
#include <mysdl/highscores.h>
#include <mysdl/files.h>
#include <mysdl/framework.h>
#include <mysdl/mysdl.h>
#include <mysdl/resources.h>
#include <mysdl/tools.h>
#include <mysdl/xml.h>

#include <mysdl/audio/core.h>
#include <mysdl/audio/music.h>
#include <mysdl/audio/resources.h>
#include <mysdl/audio/sources.h>

#include <mysdl/gfx/animations.h>
#include <mysdl/gfx/fonts.h>
#include <mysdl/gfx/images.h>

#include <mysdl/gui/containers.h>
#include <mysdl/gui/core.h>
#include <mysdl/gui/standard.h>
#include <mysdl/gui/text.h>
#include <mysdl/gui/slider.h>

#include <mysdl/particles/starfield.h>

#ifdef linux
	#define DEF_FULLSCREEN false
#else
	#define DEF_FULLSCREEN true
#endif

#define RES_X 800
#define RES_Y 600

namespace maxfighter
{
	/** Debug categories. */
	enum Debug {dbgCore=1,dbgGame,dbgResources};

	/** Global input event IDs. */
	enum CtrlGlobal {cgExit=1,
									 cgOk=2,
									 cgToggleFullscreen=3,
									 cgScreenshot=4,
									 cgPause=5};

	/** Menu input event IDs. */
	enum CtrlMenu 	{cmSinglePlayerA=100,
									 cmSinglePlayerB=101,
									 cmMultiplayer=102,
									 cmSettings=103,
									 cmCredits=104,
									 cmControlsSinglePlayer=120,
									 cmControlsMultiplayer=121,
									 cmControlsKey=122,
									 cmGetKey=123,
									 cmGetJoyButton=124,
									 cmGetJoyAxis=125,
									 cmSoundSettings=126,
									 cmVolumeSounds=127,
									 cmVolumeMusic=128};

	/** Game input event IDs. */
	enum CtrlGame 	{cgContinue=200,
									 cgQuit=201,
									 cgPlayerX=202,
									 cgPlayerY=203,
									 cgFire=204,
									 cgShield=205,
									 cgTimeWarp=206};

	/**
	 * Game configuration.
	 */
	class Settings
	{
		public:
			/** Horizontal resolution. Should not be changed in the current
			 *  implementation. */
			int xRes;

			/** Horizontal resolution. Should not be changed in the current
		 	 *  implementation. */
			int yRes;

			/** Whether the game currently runs in fullscreen mode. */
			bool fullscreen;

			/** Sound effects volume. At volume 0, no sounds are played. */
			int volumeSfx;

			/** Music volume. At volum 0, no music is played. */
			int volumeMusic;

			/**
			 * Constructor.
			 */
			Settings();
	};

	/**
	 * Callback for gui events.
	 */
	class GUICallback : public mysdl::gui::GUICallback
	{
		public:
			/**
			 * Plays a sound effects when the user moves between menu items.
			 */
			virtual void onMove();

			/**
			 * Plays a different sound effect it the user confirms/selects
			 * a menu item.
			 */
			virtual void onConfirm();
	};

	/**
	 * An entry in the highscore table, consisting of a name and a score.
	 */
	class HighscoreEntry
	{
		public:
			/** Name. */
			std::string name;

			/** Score. */
			unsigned int score;

			/**
			 * Constructor.
			 */
			HighscoreEntry();

			/**
			 * Constructor which initialises the new entry with a name and
			 * score.
			 */
			HighscoreEntry(std::string name,unsigned int score);

			/**
			 * Compares two highscore entries and determines whether this
			 * one has the higher score.
			 */
			bool operator>(const HighscoreEntry &e) const;

			/**
			 * Reads the contents of this highscore entry from a DOM node.
			 */
			void getFromNode(mysdl::xml::Node *node);

			/**
			 * Puts the contents of this highscore entry in a DOM node.
			 */
			void putInNode(mysdl::xml::Node *node);
	};

	/**
	 * A highscore table consists of ten highscore entries.
	 */
	typedef mysdl::HighscoreTable<HighscoreEntry,10> HighscoreTable;

	/**
	 * The complete set of highscores for the game. There are separate
	 * highscore tables for the three diferent game modes.
	 */
	class Highscores
	{
		public:
			/** Highscores for Single Player Mode A. */
			HighscoreTable singleA;

			/** Highscores for Single Player Mode B. */
			HighscoreTable singleB;

			/** Highscores for Multiplayer Mode. */
			HighscoreTable multi;

			/**
			 * Clears all highscore tables.
			 */
			void clear();

			/**
			 * Reads the highscore tables from a DOM node.
			 */
			void getFromNode(const mysdl::xml::Node *root);

			/**
			 * Creates a DOM tree containing the highscore tables and returns
			 * the root node.
			 */
			mysdl::xml::Node *buildDomTree();

			/**
			 * Loads the highscore tables from the specified file.
			 */
			void load(std::string filename);

			/**
			 * Writes the highscore tables to the specified file.
			 */
			void save(std::string filename);
	};

	/**
	 * A common base class for all game state classes (Menu and Game).
	 * This provides global event handling and global drawing methods.
	 */
	class GameState : public mysdl::ApplicationState
	{
		protected:
			/** Current frames per second value, updated at the beginning of
			 *  the doPerform() loop. */
			int lastFps;

			/** Whether the game is currently paused. In paused mode, nothing
			 *  happens until the player unpauses the game. */
			bool paused;

			/** When editing a set of controls, this holds the original
			 *  controls evaluator. It will be replaced only when the user
			 *  confirms the new settings. */
			mysdl::InputEvaluator *originalEvaluator;

			/** When editing a set of controls, this holds a clone of the
			 *  original controls evaluator. This will store the updated
			 *  data while editing the control set. The original evaluator
			 *  will be replaced by this data when the user confirms the
			 *  new settings. */
			mysdl::InputEvaluator *editEvaluator;

			/** When editing a specific control, this holds a reference to
			 *  the trigger for which a new key or joystick button/axis is
			 *  requested. */
			mysdl::Trigger *editTrigger;

			/** Whether any controls have been changed since the user entered
			 *  the control settings menu. */
			bool controlsChanged;

			/** Whether the set of controls currently being edited has been
			 *  changed by the user. */
			bool currentControlsChanged;

			/** The currently active controls button. Used to reactivate the
			 *  button after the user has assigned a new key/joystick to the
			 *  control. */
			mysdl::gui::Component *currentControlsButton;
			
			/** The value of the key/joy button/axis last pressed when
			 *  prompted for a new control key. This is used to decide between
			 *  negative and positive directions of the player's axes. */
			GLfloat newKeyValue;

			/** The current "angle" of the highscore position higlight.
			 *  Player's positions are highlighted by flashing their entry.
			 *  The current intensity of the colour for the flashing entry is
			 *  the sinus of this value. The angle must be between 0 and
			 *  Pi. */
			GLfloat highscoreHighlightAngle;

		public:
			/**
			 * Constructor.
			 */
			GameState();

			/**
			 * Handles global input events (e.g. fullscreen toggling and
			 * menu entries available from both the title and in game menus.
			 */
			virtual void handleEvent(const mysdl::Trigger *e);

			/**
			 * Performs global main loop tasks. This takes care of updating
			 * the current FPS value, updating sounds, input queues and the
			 * menu (all stage 0) as well as deleting obsolete sprites and
			 * updating the starfield (stage 2).
			 */
			virtual void doPerform(int stage,Uint32 ticks);

			/**
			 * Stage 0: Clears the screen and draws the starfield.
			 * Stage 2: Draws the GUI (if visible), the FPS, delays the game
			 *          to prevent overusing the CPU (TODO: do we really want to
			 *          do this HERE??) and flips the screen.
			 */
			virtual void doDraw(int stage);

			/**
			 * Starts the music for the current game state. This must be
			 * implemented in derived classes.
			 */
			virtual void startMusic()=0;

			/**
			 * Closes the currently active window.
			 *
			 * @param confirm true if the user selected "ok", false if he
			 *                selected "cancel".
			 */
			virtual bool closeWindows(bool confirm);

			/**
			 * Shows the sound settings window.
			 */
			virtual void showWindowSoundSettings();

			/**
			 * Closes the sound settings window.
			 *
			 * @param confirm true if the user selected "ok", false if he
			 *                selected "cancel". If this is true, the new
			 *                values are persisted, otherwise the previous
			 *                values are restored.
			 */
			virtual void closeWindowSoundSettings(bool confirm);

			/**
			 * Shows the controls window.
			 *
			 * @param title     Title caption for the window.
			 * @param evaluator The input evaluator that will be edited in
			 *                  this window.
			 */
			virtual void showWindowControls(const std::string &title,mysdl::InputEvaluator *evaluator);

			/**
			 * Closes the controls window.
			 *
			 * @param confirm true if the user selected "ok", false if he
			 *                selected "cancel". If this is true, the new
			 *                values are persisted, otherwise the previous
			 *                values are restored.
			 */
			virtual void closeWindowControls(bool confirm);

			/**
			 * Updates the labels in the controls window according to the
			 * current contents of the evaluator being edited.
			 */
			virtual void updateKeyLabels();

			/**
			 * Displays a highscore table.
			 *
			 * @param xOffset    X offset of the highscore table's center.
			 * @param yOffset    Y offset of the highscore table's center.
			 * @param highscores Highscore table to be displayed.
			 * @param posRed     Index of the entry that should be flashign
			 *                   between red and white (-1 for none).
			 * @param posBlue    Index of the entry that should be flashign
			 *                   between blue and white (-1 for none).
			 */
			virtual void displayHighscores(GLfloat xOffset,GLfloat yOffset,HighscoreTable *highscores,int posRed,int posBlue);
	};

	/**
	 * Initialises and stores all resources required by the game.
	 */
	class Resources : public mysdl::ResourceManager,public mysdl::BackgroundTaskCallback
	{
		public:
			/** Game configuration. */
			Settings settings;

			/** Whether the game has finished loading all images. */
			bool imagesLoaded;

			/** The viewport representing the playfield screen. */
			mysdl::Viewport vpPlayfield;

			/** The context for drawing the game scene. */
			mysdl::RenderContext2D contextGame;

			/** The context for drawing the game HUD. */
			mysdl::RenderContext2D contextHud;

			/** The context for drawing the GUI. */
			mysdl::RenderContext2D contextMenu;

			/** All fonts used by the game. */
			mysdl::fonts::FontBank *fonts;

			/** The standard font. */
			mysdl::fonts::Font *fontStandard;

			/** Images used in the game. */
			mysdl::gfx::ImageBank *imagesGame;

			/** Animations used in the game. */
			mysdl::gfx::AnimationBank *animationsGame;

			/** Images used to draw the HUD. */
			mysdl::gfx::ImageBank *imagesHud;

			/** Images used to draw the GUI. */
			mysdl::gfx::ImageBank *imagesGui;

			/** Menu GUI. */
			mysdl::gui::GUI *gui;

			/** The first two joysticks (if available) are automatically
			 *  opened and can be used to play the game (TODO: more?) */
			mysdl::Joystick *joystick[2];

			/** Evaluator for global events. */
			mysdl::InputEvaluator *evaluatorCore;

			/** Evaluator for menu events. */
			mysdl::InputEvaluator *evaluatorMenu;

			/** Evaluator for global game events. */
			mysdl::InputEvaluator *evaluatorGameCore;

			/** Evaluator for single player events. */
			mysdl::InputEvaluator *evaluatorPlayerSingle[1];

			/** Two evalators, one for each player's events in a mutliplayer
			 *  game. */
			mysdl::InputEvaluator *evaluatorPlayerMulti[2];

			/** Control queue. */
			mysdl::ControlQueue *queue;

			/** White, a common colour used throughout the game. */
			mysdl::Color colorWhite;

			/** Properties for the starfield. */
			mysdl::particles::StarfieldProperties starfieldProperties;

			/** The starfield, always displayed in the background during the
			 *  entire game. */
			mysdl::particles::Starfield *starfield;

			/** All cached sound effects for the menu and game. */
			mysdl::audio::SampleBank *samples;

			/** All streamed music samples for the menu and game. */
			mysdl::audio::SampleBank *music;

			/** References to the cached sound effect samples. */
			mysdl::audio::Sample /* *sampleBeep,*/ *sampleExplosion,*sampleLaser,*samplePing,*sampleExtra,*sampleTimeWarp,*sampleMenuMove,*sampleMenuClick;

			/** References to the streamed music samples. */
			mysdl::audio::StreamedSample *musicTitleIntro,*musicTitleLoop;
			mysdl::audio::StreamedSample *musicGameAIntro,*musicGameALoop,*musicGameBIntro,*musicGameBLoop,*musicDeath;
			mysdl::audio::StreamedSample *musicHighscoresIntro,*musicHighscoresLoop,*musicGameOverIntro,*musicGameOverLoop;

			/** Sources for general, location-less sound effects. */
			mysdl::audio::SourcePool *sourcesGeneral;

			/** Sources for ingame sound effects which usually have a
			 *  specific position. */
			mysdl::audio::SourcePool *sourcesExplosions;

			/** Music manager. */
			mysdl::audio::MusicManager *musicManager;

			/** All highscore tables for the game. */
			Highscores highscores;

			/** Logo image for the title menu. */
			mysdl::gfx::Image *imgLogo;

			/** References to the various images and animations used during the game. */
			mysdl::gfx::Animation *animPlayer[2];
			mysdl::gfx::Image *imgLaserRay,*imgLaserRound,*imgLaserFoe;
			mysdl::gfx::Animation *animEngine,*animEngineStutter,*animShield,*animAsteroidSmall,*animAsteroidLarge,*animCrystal;
			mysdl::gfx::Animation *animContainer[3];
			mysdl::gfx::Animation *animExtraFireRate,*animExtraLaserType,*animExtraSpeedup,*animExtraShield,*animExtraScore;
			mysdl::gfx::Animation *animExtraTimeWarp;
			mysdl::gfx::Animation *animFoeLeft,*animFoeRight;
			mysdl::gfx::Animation *animExplosionFast,*animExplosionSlow;

			/** References to the images used to draw the HUD. */
			mysdl::gfx::Image *imgHudFull;
			mysdl::gfx::Image *imgHudShield;
			mysdl::gfx::Image *imgHudSpeed[5];
			mysdl::gfx::Image *imgHudFire[5][6];
			mysdl::gfx::Image *imgHudTimewarp[2];

			/** A MySDL image loading thread. */
			mysdl::gfx::ImageLoader *imageLoader;

			/** References to various GUI widgets. */
			mysdl::gui::Pane *paneMain,*paneSettings,*paneControls,*paneCredits,*paneGetKey,*paneSoundSettings,*paneEnterName,*paneInGame,*paneQuit;
			mysdl::gui::Button *btnQuitNo,*btnQuit,*btnSinglePlayerA,*btnSinglePlayerB,*btnMultiplayer,*btnSettings,*btnCredits;
			mysdl::gui::Button *btnControlsSinglePlayer,*btnControlsMultiplayer[2],*btnSettingsDone;
			mysdl::gui::Button *btnSoundSettings,*btnSoundSettingsOk;
			mysdl::gui::Checkbox *cbFullscreen;
			mysdl::gui::Button *btnControlsPlayer1,*btnControlsPlayer2,*btnControlsLeft,*btnControlsRight,*btnControlsFire;
			mysdl::gui::Label *labelControlsHeader;
			mysdl::gui::Label *labelControlsLeft,*labelControlsRight,*labelControlsUp,*labelControlsDown;
			mysdl::gui::Label *labelControlsLeftJoy,*labelControlsRightJoy,*labelControlsUpJoy,*labelControlsDownJoy;
			mysdl::gui::Label *labelControlsFire,*labelControlsShield,*labelControlsTimeWarp;
			mysdl::gui::Label *labelControlsFireJoy,*labelControlsShieldJoy,*labelControlsTimeWarpJoy,*labelControlsPauseJoy;
			mysdl::gui::Label *labelGetKeyTriggerName,*labelGetKeyHeader,*labelGetKeyHeaderAlternate;
			mysdl::gui::Label *labelHeaderVersion;
			mysdl::gui::Button *btnContinue,*btnInGameSoundSettings,*btnInGameControlsSinglePlayer,*btnInGameControlsMultiplayer[2];
			mysdl::gui::Checkbox *cbInGameFullscreen;
			mysdl::gui::Slider<int> *sVolumeSounds,*sVolumeMusic;
			mysdl::gui::Label *lVolumeSounds,*lVolumeMusic;
			mysdl::gui::LabelStyle *lsTextfieldRed,*lsTextfieldBlue;
			mysdl::gui::InputCaptor *icGetKey;

			/** Current location of the logo in the title menu. */
			mysdl::basic_Vector2D<GLfloat> logoPosition;

			/** Current colour of the logo in the title menu. */
			mysdl::Color logoColor;

			/** Current sound effects volume. This may differ from the volume
			 *  specified in the settings only if the user is currently
			 *  changing sound settings. */
			int currentVolumeSfx;

			/** Current msic volume. This may differ from the volume
			 *  specified in the settings only if the user is currently
			 *  changing sound settings. */
			int currentVolumeMusic;

			/**
			 * Constructor.
			 */
			Resources();

			/**
			 * Destructor.
			 */
			~Resources();

			/**
			 * Sets/changes the resolution and fullscreen mode. Once the
			 * resolution has been set for the first time, Linux can toggle
			 * between windowed and fullscreen mode on the fly, while Windows
			 * has to re-load all images. This is automaticall taken care of
			 * by this method.
			 */
			void setResolution(int width,int height,int depth,bool fullscreen);

			/**
			 * Saves the highscores.
			 */
			void saveHighscores();

			/**
			 * Initialises global resources.
			 */
			void init();

			/**
			 * Initialises default highscores.
			 */
			void initHighscores();

			/**
			 * Initialises menu resources.
			 */
			void initMenu();

			/**
			 * Uninitialises menu resources.
			 */
			void uninitMenu();

			/**
			 * Initialises game resources.
			 */
			void initGame();

			/**
			 * Uninitialises game resources.
			 */
			void uninitGame();

			/**
			 * Triggered when the background image loader thread starts.
			 */
			void onStart(mysdl::BackgroundTask *task);

			/**
			 * Triggered when the background image loader thread finished.
			 * At this point, we switch from the title screen to the title
			 * menu and allow user input.
			 */
			void onFinish(mysdl::BackgroundTask *task);

			/**
			 * Loads game settings.
			 */
			void loadSettings();

			/**
			 * Saves game settings.
			 */
			void saveSettings();

			/**
			 * Updates all "toggle fullscreen" GUI buttons to show the
			 * current fullscreen state.
			 */
			void updateFullscreenButtons();

			/**
			 * Sets the current volume for sound effects. Does NOT affect
			 * the settings.
			 */
			void setVolumeSfx(int volume);

			/**
			 * Sets the current volume for music. Does NOT affect the
			 * settings.
			 */
			void setVolumeMusic(int volume);
	};

	/** Connected to the GUI to play sounds for menu events. */
	extern GUICallback callback;

	extern mysdl::SysDirManager dirManager;
};

#endif
