/**
 * Max Fighter
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://maxfighter.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
#ifndef MAXFIGHTER_SETTINGS_H
#define MAXFIGHTER_SETTINGS_H

#include <mysdl/mysdl.h>

#include "resources.h"

namespace maxfighter
{
	/**
	 * This class stores game specification, i.e. values that affect
	 * gameplay and menu behaviour. This is done so that these values
	 * can easily be changed in a global location, without having to
	 * recompile the entire project.
	 */
	class Specs
	{
		public:
			/** The maximum number of milliseconds that may pass between two
			 *  frames. If two frames are further apart than the amount of
			 *  time specified here, the game pretends that the surplus time
			 *  never happened. */
			Uint32 maxTicksPerFrame;

			/** The desired number of milliseconds between two frames. If
			 *  the system is fast enough to draw frames more frequently, the
			 *  game will yield control to the CPU during the surplus
			 *  time. */
			Uint32 desiredTicks;

			/** FIXME: It seems all this does is reduce the desiredTicks
			 *  value, so what good is it really? (Besides, it is 0) */
			Uint32 desiredTicksEpsilon;

			/** Number of pixels between the allowed range for the players'
			 *  hotspots and the left and right edge of the screen. */
			GLfloat playfieldBorderX;

			/** Initial Y position for players. (TODO: ignored?) */
			GLfloat startY;

			/** Stasndard Y position for players. */
			GLfloat standardY;

			/** Topmost Y position for players in Game Mode A. */
			GLfloat maxYModeA;

			/** Topmost Y position for players in Game Mode B. */
			GLfloat maxYModeB;

			/** Minimum X distance between two players. */
			int minPlayerDistanceX;

			/** Minimum Y distance between two players. */
			int minPlayerDistanceY;

			/** Vertical speed of a player in relation to his horizontal
			 *  speed (1:n). */
			GLfloat playerSpeedFactorY;

			/** FIXME: This is an additional factor to scale the velocity
			 *  vectors of containers? Why? */
			GLfloat baseSpeed;

			/** The speed factor for a player who has no speedup extras. */
			GLfloat speedStart;

			/** The amount by which a player's speed factor is increased each
			 *  time he collects a speedup extra. */
			GLfloat speedIncrement;

			/** First level in which shield extras may appear. */
			int minLevelForShield;

			/** First level in which weapon extras (i.e. timewarps) may
			 *  appear. */
			int minLevelForWeapons;

			/** The horizontal offset of the two smaller asteroids into which
			 *  an exploding large asteroid splits, relative to the original
			 *  speed vector of the large asteroid. */
			GLfloat splitSpeedX;

			/** Number of stars in the starfield. */
			Uint32 starCount;

			/** Factor for calculating the starfield speed from the current
			 *  game speed. */
			GLfloat starSpeedFactor;

			/** Maximum number of speedup extras a player can effectively
			 * collect. */
			int speedupCountMax;

			/** Maximum number of fire rate extras a player can effectively
			 * collect. */
			int fireRateCountMax;

			/** Maximum number of laser type extras a player can effectively
			 * collect. */
			int laserLevelMax;

			/** Factor for calculating a container's speed from the current
			 *  game speed. */
			GLfloat containerSpeedFactor;

			/** Factor for calculating an extra's speed from the current
			 *  game speed. */
			GLfloat extraSpeedFactor;

			/** The amount by which a fading extra fades each millisecond. */
			GLfloat extraFadeFactor;

			/** Number of frames in the player tilt animation. */
 			Uint32 playerTiltAnimFrames;

			/** Amount by which a player's shield energy is increased when
			 *  collecting a shield upgrade. */
			GLfloat shieldEnergyPacket;

			/** Maximum shield energy for players. */
			GLfloat shieldEnergyMax;

			/** Maximum number of timewarp items a player can effectively
			 *  collect. */
			int timeWarpLevelMax;

			/** The amount by which a player's shield fades in per
			 *  millisecond. */
			GLfloat shieldFadeInFactor;

			/** The amount by which a player's shield fades out per
			 *  millisecond. */
			GLfloat shieldFadeOutFactor;

			/** The amount by which a player's shield fades out per
			 *  millisecond in collapse mode. */
			GLfloat shieldCollapseFactor;

			/** The factor by which a tick is scaled in timewarp mode. */
			GLfloat timeWarpFactor;

			/** The amount by which the tick scaling approaches the
			 *  timewarp factor per millisecond. */
			GLfloat timeWarpStartSpeed;

			/** The amount by which the tick scaling approaches the
			 *  standard factor per millisecond (after timewarp has
			 *  finished). */
			GLfloat timeWarpStopSpeed;

			/** Number of milliseconds for which timewarp is active. */
			Uint32 timeWarpDuration;

			/** Whenever a player has collected this amount of points, a
			 *  new container is queued. */
 			Uint32 containerScoreStep;

			/** Starfield speed in menu mode. */
			GLfloat starfieldMenuSpeed;

			/** The value by which the starfield speed may change in menu
			 *  mode to approach the predefined speed. */
			GLfloat starfieldMenuSpeedStep;

			/** The amount by which an extra's speed vector is scaled :*/
			GLfloat extraVelocityFactor;

			/** Number of milliseconds between two score extras in Game
			 *  Mode B. */
			Uint32 extraIntervalScore;

			/** Number of milliseconds between two speedup extras in Game
			 *  Mode B. */
			Uint32 extraIntervalExtra;

			/** The amount by which the player's score is increased when
			 *  the speed is increased in B mode. */
			Uint32 scoreIncrease;

			/** The amount by which the game speed is increased every time
			 *  the speed is increased in B mode. */
			GLfloat speedIncreaseB;

			/** The amount by which the asteroid interval is decreased in
			 *  B mode. */
			Uint32 asteroidIntervalDecreaseB;

			/** The minimum asteroid interval (in milliseconds) in B mode. */
			Uint32 asteroidIntervalMinB;

			/** The chance (between 0 and 1) of an asteroid being a large
			 *  asteroid in B mode. */
			GLfloat chanceLargeAsteroidB;

			/** The number of milliseconds after which a lost extra "expires"
			 *  and cannot be recollected. */
			Uint32 extraReviveTicks;

			/** FIXME: this is 0... it has no effect... what is it for?? */
			int extraContainerLimit;

			/** Initial asteroid positions and speeds are calculated so that
			 *  the asteroid will be visible on screen at this Y position. */
			GLfloat asteroidTargetY;

			/** Number of ticks between two HUD states when the HUD is in
			 *  flashing mode. */
			GLfloat hudSwitchTime;

			/** Number of switches between HUD states during flasing mode. */
			int hudSwitchCount;

			/** The amount of energy an active shield consumes each
			 *  millisecond. */
			GLfloat shieldConsumption;

			/** Number of milliseconds until a new game can be started after
			 *  the highscores are displayed. */
			GLfloat restartDelay;

			/** Number of points awarded for a crystal. */
			Uint32 scoreCrystal;

			/** Number of points awarded for destroying a foe. */
			Uint32 scoreFoe;

			/** Number of points awarded for destroying a small asteroid. */
			Uint32 scoreAsteroidSmall;

			/** Number of points awarded for destroying a large asteroid. */
			Uint32 scoreAsteroidLarge;

			/** Maxmimum amount of bonus points. */
			Uint32 scoreBonus;

			/** The first level in which a foe appears. */
			Uint32 foeFirstLevel;

			/** The amount by which a foe's velocity level is scaled. */
			GLfloat foeSpeed;

			/** The horizontal component of a foe's laser velocity vector. */
			GLfloat foeLaserSpeed;

			/** The factor by which the vertical component of a foe's laser
			 *  velocity vector is scaled. */
			GLfloat foeLaserSpeedVertical;

			/** Topmost Y position of a foe. */
			GLfloat foeTop;

			/** Bottommost Y position of a foe. */
			GLfloat foeFloor;

			/** The amount by which a foe's laser shot fades out each
			 *  millisecond. TODO: TTL? */
			GLfloat foeLaserFadeOutSpeed;

			/** Initial game speed in mode B. */
			GLfloat startSpeedModeB;

			/** Initial asteroid interval in mode B. */
			Uint32 startAsteroidIntervalModeB;

			/** Number of milliseconds until the menu automatically switches
			 *  away from menu mode and into the first highscore mode. */
			Uint32 menuSwitchDuration;

			/** Number of milliseconds until the menu automatically switches
			 *  away from any highscore mode. */
			Uint32 menuHighscoreDuration;

			/** Number of milliseconds until the menu automatically switches
			 *  away from help mode. */
			Uint32 menuHelpStepDuration;

			/** Center position of the HUD for player 1. */
			mysdl::Vector2D hudPosPlayer1;

			/** Center position of the HUD for player 2. */
			mysdl::Vector2D hudPosPlayer2;

			/** Background colour for drawing the HUD. This colour is used
			 *  to draw the entire HUD regardless of the player's current
			 *  state. */
			mysdl::Color colorHudBack;

			/** Foreground colour for drawing the HUD. This colour is used
			 *  to draw the HUD representing the player's state over a full
			 *  HUD previously drawn in the background colour. */
			mysdl::Color colorHudFront;

			/**
			 * Contructor, initialises all values.
			 */
			Specs();

			/**
			 * Gets the number of milliseconds between spawning two asteroids
			 * in the specified level.
			 */
			Uint32 getAsteroidInterval(Uint32 level);

			/**
			 * Gets a random velocity for an asteroid in the specified level.
			 */
			mysdl::Vector2D getAsteroidVelocity(GLfloat speed);

			/**
			 * Gets the game speed for the specified level.
			 */
			GLfloat getSpeedForLevel(Uint32 level);

			/**
			 * Gets the number of asteroids that must be destroyed in the
			 * specified level.
			 */
			Uint32 getKillsForLevel(Uint32 level);

			/**
			 * Gets the number of milliseconds until the next speedup when
			 * the speed has just been set to the specified value in Game
			 * Mode B.
			 */
			Uint32 speedupInterval(GLfloat speed);

			/**
			 * Gets the number of milliseconds between two laser shots
			 * at the specified laser rate level.
			 */
			GLfloat getLaserDelay(short int level);

			/**
			 * Gets the velocity for laser shots at the specified laser rate
			 * level.
			 */
			GLfloat getLaserSpeed(short int level);

			/**
			 * Gets the chance (between 0.0 and 1.0) of an asteroid being a
			 * large asteroid in the specified level.
			 */
			GLfloat getChanceLargeAsteroid(short int level);

			/**
			 * Gets the chance (between 0.0and 1.0) of an asteroid having
			 * an initial energy value greater than 0 in the specified
			 * level.
			 */
			GLfloat getChanceAsteroidEnergy(short int level);

			/**
			 * Gets the maximum asteroid energy value in the specified level.
			 */
			Uint32 getMaxAsteroidEnergy(short int level);

			/**
			 * Gets the number of milliseconds between two laser shots fired
			 * by a foe in the specified level.
			 */
			Uint32 getFoeFireDelay(short int level);

			/**
			 * Gets the (random) number of shots per volley fired by a foe in
			 * the specified level.
			 */
			Uint32 getFoeVolleyCount(short int level);

			/**
			 * Gets the number of milliseconds between two laser shots of
			 * the same volley fired by a foe in the specified level.
			 */
			Uint32 getFoeVolleyDelay(short int level);

			/**
			 * Gets the initial hull energy for a foe in the specified level.
			 */
			Uint32 getFoeHull(short int level);

			/**
			 * Gets the acceleration factor for a foe in the specified level.
			 */
			GLfloat getFoeAcceleration(short int level);

			/**
			 * Gets the factor which defines how quickly a foe in the
			 * specified level changes its speed depending on the distance
			 * to the targetted player.
			 */
			GLfloat getFoeDistanceSpeedFactor(short int level);

			/**
			 * Gets the number of milliseconds a shot fired by a foe in the
			 * specified level "lives".
			 */
			Uint32 getFoeLaserTtl(short int level);

			/**
			 * Gets the chance (1:n) of a foe entering kamikaze mode at any
			 * moment (millisecond?) in the specified level.
			 */
			int getFoeKamikazeChance(short int level);

			/**
			 * Gets the chance (1:n) of a foe entering flooring mode at any
			 * moment (millisecond?) in the specified level.
			 */
			int getFoeFloorChance(short int level);

			/**
			 * Gets the (random) number of milliseconds a foe in the
			 * specified level will remain in flooring mode.
			 */
			Uint32 getFoeFloorDuration(short int level);
	};

	/** Global (constant) instance containing game specifications. */
	extern Specs specs;
};

#endif
